from __future__ import annotations

import importlib
from typing import TYPE_CHECKING, Any, get_args, overload

import numpy as np
import pandas as pd
from django.db import models
from django.db.models import CASCADE, PROTECT
from django.db.models.query_utils import DeferredAttribute
from django.db.utils import IntegrityError
from lamin_utils import logger
from lamindb_setup._init_instance import get_schema_module_name
from lamindb_setup.core import deprecated
from lamindb_setup.core.hashing import HASH_LENGTH, hash_dict, hash_string
from lamindb_setup.errors import (
    MODULE_WASNT_CONFIGURED_MESSAGE_TEMPLATE,
    ModuleWasntConfigured,
)
from pandas.api.types import CategoricalDtype, is_string_dtype
from pandas.core.dtypes.base import ExtensionDtype

from lamindb.base.fields import (
    BooleanField,
    CharField,
    ForeignKey,
    JSONField,
    TextField,
)
from lamindb.base.types import Dtype, FieldAttr
from lamindb.errors import DoesNotExist, FieldValidationError, ValidationError

from ..base.ids import base62_12
from ._relations import dict_module_name_to_model_name
from .can_curate import CanCurate
from .query_set import SQLRecordList
from .run import (
    TracksRun,
    TracksUpdates,
)
from .sqlrecord import BaseSQLRecord, Registry, SQLRecord, _get_record_kwargs

if TYPE_CHECKING:
    from collections.abc import Iterable

    from .schema import Schema

FEATURE_DTYPES = set(get_args(Dtype))


def parse_dtype(dtype_str: str, is_param: bool = False) -> list[dict[str, Any]]:
    """Parses feature data type string into a structured list of components."""
    from .artifact import Artifact

    allowed_dtypes = FEATURE_DTYPES
    if is_param:
        allowed_dtypes.add("dict")

    # Handle list[...] types
    if dtype_str.startswith("list[") and dtype_str.endswith("]"):
        inner_dtype_str = dtype_str[5:-1]  # Remove "list[" and "]"
        # Recursively parse the inner type
        inner_result = parse_dtype(inner_dtype_str, is_param)
        # Add "list": True to each component
        for component in inner_result:
            if isinstance(component, dict):
                component["list"] = True  # type: ignore
        return inner_result

    is_composed_cat = dtype_str.startswith("cat[") and dtype_str.endswith("]")
    result = []
    if is_composed_cat:
        related_registries = dict_module_name_to_model_name(Artifact)
        registries_str = dtype_str.replace("cat[", "")[:-1]  # strip last ]
        if registries_str != "":
            registry_str_list = registries_str.split("|")
            for cat_single_dtype_str in registry_str_list:
                single_result = parse_cat_dtype(
                    cat_single_dtype_str, related_registries
                )
                result.append(single_result)
    elif dtype_str not in allowed_dtypes:
        raise ValueError(
            f"dtype is '{dtype_str}' but has to be one of {FEATURE_DTYPES}!"
        )
    return result


def parse_cat_dtype(
    dtype_str: str,
    related_registries: dict[str, SQLRecord] | None = None,
    is_itype: bool = False,
) -> dict[str, Any]:
    """Parses a categorical dtype string into its components (registry, field, subtypes)."""
    from .artifact import Artifact

    assert isinstance(dtype_str, str)  # noqa: S101
    if related_registries is None:
        related_registries = dict_module_name_to_model_name(Artifact)

    # Parse the string considering nested brackets
    parsed = parse_nested_brackets(dtype_str)

    registry_str = parsed["registry"]
    sub_type_str = parsed["subtype"]
    field_str = parsed["field"]

    if not is_itype:
        if registry_str not in related_registries:
            raise ValidationError(
                f"'{registry_str}' is an invalid dtype, has to be registry, e.g. ULabel or bionty.CellType"
            )
        registry = related_registries[registry_str]
    else:
        if "." in registry_str:
            registry_str_split = registry_str.split(".")
            assert len(registry_str_split) == 2, registry_str  # noqa: S101
            module_name_attempt, class_name = registry_str_split
            module_name = get_schema_module_name(
                module_name_attempt, raise_import_error=False
            )
            if module_name is None:
                raise ModuleWasntConfigured(
                    MODULE_WASNT_CONFIGURED_MESSAGE_TEMPLATE.format(module_name_attempt)
                )
        else:
            module_name, class_name = "lamindb", registry_str
        module = importlib.import_module(module_name)
        registry = getattr(module, class_name)

    if sub_type_str != "":
        pass
        # validate that the subtype is a record in the registry with is_type = True
    if field_str != "":
        pass
        # validate that field_str is an actual field of the module
    else:
        field_str = registry._name_field if hasattr(registry, "_name_field") else "name"

    result = {
        "registry": registry,  # should be typed as CanCurate
        "registry_str": registry_str,
        "subtype_str": sub_type_str,
        "field_str": field_str,
        "field": getattr(registry, field_str),
    }

    # Add nested subtype information if present
    if parsed.get("nested_subtypes"):
        result["nested_subtypes"] = parsed["nested_subtypes"]

    return result


def parse_nested_brackets(dtype_str: str) -> dict[str, str]:
    """Parse dtype string with potentially nested brackets.

    Examples:
        "A" -> {"registry": "A", "subtype": "", "field": ""}
        "A.field" -> {"registry": "A", "subtype": "", "field": "field"}
        "A[B]" -> {"registry": "A", "subtype": "B", "field": ""}
        "A[B].field" -> {"registry": "A", "subtype": "B", "field": "field"}
        "A[B[C]]" -> {"registry": "A", "subtype": "B[C]", "field": "", "nested_subtypes": ["B", "C"]}
        "A[B[C]].field" -> {"registry": "A", "subtype": "B[C]", "field": "field", "nested_subtypes": ["B", "C"]}

    Args:
        dtype_str: The dtype string to parse

    Returns:
        Dictionary with parsed components
    """
    if "[" not in dtype_str:
        # No brackets - handle simple cases like "A" or "A.field"
        if "." in dtype_str:
            parts = dtype_str.split(".")
            if len(parts) == 2 and parts[1][0].isupper():
                # bionty.CellType
                return {"registry": dtype_str, "subtype": "", "field": ""}
            elif len(parts) == 3:
                # bionty.CellType.name
                return {
                    "registry": f"{parts[0]}.{parts[1]}",
                    "subtype": "",
                    "field": parts[2],
                }
            else:
                # ULabel.name
                return {"registry": parts[0], "subtype": "", "field": parts[1]}
        else:
            # Simple registry name
            return {"registry": dtype_str, "subtype": "", "field": ""}

    # Find the first opening bracket
    first_bracket = dtype_str.index("[")
    # Handle case where registry_part contains a field (e.g., "bionty.Gene.ensembl_gene_id[filters]")
    registry_and_field = dtype_str[:first_bracket]
    if "." in registry_and_field:
        parts = registry_and_field.split(".")
        if len(parts) == 3:
            registry_part = f"{parts[0]}.{parts[1]}"
            pre_bracket_field = parts[2]
        else:
            registry_part = registry_and_field
            pre_bracket_field = ""
    else:
        registry_part = registry_and_field
        pre_bracket_field = ""

    # Find the matching closing bracket for the first opening bracket
    bracket_count = 0
    closing_bracket_pos = -1

    for i in range(first_bracket, len(dtype_str)):
        if dtype_str[i] == "[":
            bracket_count += 1
        elif dtype_str[i] == "]":
            bracket_count -= 1
            if bracket_count == 0:
                closing_bracket_pos = i
                break

    if closing_bracket_pos == -1:
        raise ValueError(f"Unmatched brackets in dtype string: {dtype_str}")

    # Extract subtype (everything between first [ and matching ])
    subtype_part = dtype_str[first_bracket + 1 : closing_bracket_pos]

    # Check for field after the closing bracket
    field_part = ""
    remainder = dtype_str[closing_bracket_pos + 1 :]
    if remainder.startswith("."):
        field_part = remainder[1:]  # Remove the dot

    # Use pre_bracket_field if no post_bracket field
    if not field_part and pre_bracket_field:
        field_part = pre_bracket_field

    result = {"registry": registry_part, "subtype": subtype_part, "field": field_part}

    # If subtype contains brackets, extract nested subtypes for reference
    if "[" in subtype_part:
        nested_subtypes = extract_nested_subtypes(subtype_part)
        if nested_subtypes:
            result["nested_subtypes"] = nested_subtypes  # type: ignore

    return result


def extract_nested_subtypes(subtype_str: str) -> list[str]:
    """Extract all nested subtype levels from a nested subtype string.

    Examples:
        "B[C]" -> ["B", "C"]
        "B[C[D]]" -> ["B", "C", "D"]
        "B[C[D[E]]]" -> ["B", "C", "D", "E"]

    Args:
        subtype_str: The subtype string with potential nesting

    Returns:
        List of subtype levels from outermost to innermost
    """
    subtypes = []
    current = subtype_str

    while "[" in current:
        # Find the first part before the bracket
        bracket_pos = current.index("[")
        subtypes.append(current[:bracket_pos])

        # Find the matching closing bracket
        bracket_count = 0
        closing_pos = -1

        for i in range(bracket_pos, len(current)):
            if current[i] == "[":
                bracket_count += 1
            elif current[i] == "]":
                bracket_count -= 1
                if bracket_count == 0:
                    closing_pos = i
                    break

        if closing_pos == -1:
            break

        # Move to the content inside the brackets
        current = current[bracket_pos + 1 : closing_pos]

    # Add the final innermost subtype
    if current:
        subtypes.append(current)

    return subtypes


def serialize_dtype(
    dtype: Registry
    | SQLRecord
    | FieldAttr
    | list[SQLRecord]
    | list[Registry]
    | list[str]
    | list[float]
    | str
    | type,
    is_itype: bool = False,
) -> str:
    """Converts a data type object into its string representation."""
    from .record import Record
    from .ulabel import ULabel

    # Handle generic types like list[str], list[Registry], etc.
    if hasattr(dtype, "__origin__") and dtype.__origin__ is list:
        # Get the inner type from list[T]
        inner_type = dtype.__args__[0] if dtype.__args__ else None  # type: ignore
        if inner_type is not None:
            # Recursively serialize the inner type
            inner_dtype_str = serialize_dtype(inner_type, is_itype=is_itype)
            return f"list[{inner_dtype_str}]"

    if (
        not isinstance(dtype, list)
        and hasattr(dtype, "__name__")
        and dtype.__name__ in FEATURE_DTYPES
    ):
        dtype_str = dtype.__name__
    elif dtype is dict:
        dtype_str = "dict"
    elif is_itype and isinstance(dtype, str):
        if dtype not in "Feature":
            parse_cat_dtype(
                dtype_str=dtype, is_itype=True
            )  # throws an error if invalid
        dtype_str = dtype
    elif isinstance(dtype, (ExtensionDtype, np.dtype)):
        dtype_str = serialize_pandas_dtype(dtype)
    else:
        error_message = "dtype has to be a registry, a ulabel subtype, a registry field, or a list of registries or fields, not {}"
        if isinstance(dtype, (Registry, DeferredAttribute, ULabel, Record)):
            dtype = [dtype]
        elif not isinstance(dtype, list):
            raise ValueError(error_message.format(dtype))
        dtype_str = ""
        for one_dtype in dtype:
            if not isinstance(one_dtype, (Registry, DeferredAttribute, ULabel, Record)):
                raise ValueError(error_message.format(one_dtype))
            if isinstance(one_dtype, Registry):
                dtype_str += one_dtype.__get_name_with_module__() + "|"
            elif isinstance(one_dtype, (ULabel, Record)):
                assert one_dtype.is_type, (  # noqa: S101
                    f"ulabel has to be a type if acting as dtype, {one_dtype} has `is_type` False"
                )
                if isinstance(one_dtype, ULabel):
                    dtype_str += f"ULabel[{one_dtype.name}]"
                else:
                    dtype_str += f"Record[{one_dtype.name}]"
            else:
                name = one_dtype.field.name
                field_ext = f".{name}" if name != "name" else ""
                dtype_str += (
                    one_dtype.field.model.__get_name_with_module__() + field_ext + "|"
                )
        dtype_str = dtype_str.rstrip("|")
        if not is_itype:
            dtype_str = f"cat[{dtype_str}]"
    return dtype_str


def serialize_pandas_dtype(pandas_dtype: ExtensionDtype) -> str:
    """Convert pandas ExtensionDtype to simplified string representation."""
    if is_string_dtype(pandas_dtype):
        if not isinstance(pandas_dtype, CategoricalDtype):
            dtype = "str"
        else:
            dtype = "cat"
    # there are string-like categoricals and "pure" categoricals (pd.Categorical)
    elif isinstance(pandas_dtype, CategoricalDtype):
        dtype = "cat"
    else:
        # strip precision qualifiers
        dtype = "".join(dt for dt in pandas_dtype.name if not dt.isdigit())
        if dtype == "uint":
            dtype = "int"
    if dtype.startswith("datetime"):
        dtype = dtype.split("[")[0]
    assert dtype in FEATURE_DTYPES  # noqa: S101
    return dtype


def parse_filter_string(filter_str: str) -> dict[str, tuple[str, str | None, str]]:
    """Parse comma-separated Django filter expressions into structured components.

    Args:
        filter_str: Comma-separated filters like 'name=value, relation__field=value'

    Returns:
        Dict mapping original filter key to (relation_name, field_name, value) tuple.
        For direct fields: field_name is None.
        For relations: field_name contains the lookup field.
    """
    filters = {}

    filter_parts = [part.strip() for part in filter_str.split(",")]
    for part in filter_parts:
        if "=" not in part:
            raise ValueError(f"Invalid filter expression: '{part}' (missing '=' sign)")

        key, value = part.split("=", 1)
        key = key.strip()
        value = value.strip().strip("'\"")

        if not key:
            raise ValueError(f"Invalid filter expression: '{part}' (empty key)")
        if not value:
            raise ValueError(f"Invalid filter expression: '{part}' (empty value)")

        if "__" in key:
            relation_name, field_name = key.split("__", 1)
            filters[key] = (relation_name, field_name, value)
        else:
            filters[key] = (key, None, value)

    return filters


def resolve_relation_filters(
    parsed_filters: dict[str, tuple[str, str | None, str]], registry: SQLRecord
) -> dict[str, str | SQLRecord]:
    """Resolve relation filters actual model objects.

    Args:
        parsed_filters: Django filters like output from :func:`lamindb.models.feature.parse_filter_string`
        registry: Model class to resolve relationships against

    Returns:
        Dict with resolved objects for successful relations, original values for direct fields and failed resolutions.
    """
    resolved = {}

    for filter_key, (relation_name, field_name, value) in parsed_filters.items():
        if field_name is not None:  # relation filter
            if hasattr(registry, relation_name):
                relation_field = getattr(registry, relation_name)
                if (
                    hasattr(relation_field, "field")
                    and relation_field.field.is_relation
                ):
                    try:
                        related_model = relation_field.field.related_model
                        related_obj = related_model.get(**{field_name: value})
                        resolved[relation_name] = related_obj
                        continue
                    except (DoesNotExist, AttributeError):
                        pass  # Fall back to original filter
        resolved[filter_key] = value

    return resolved


def process_init_feature_param(args, kwargs, is_param: bool = False):
    # now we proceed with the user-facing constructor
    if len(args) != 0:
        raise ValueError("Only keyword args allowed")
    name: str = kwargs.pop("name", None)
    dtype: type | str | None = kwargs.pop("dtype", None)
    is_type: bool = kwargs.pop("is_type", None)
    type_: Feature | str | None = kwargs.pop("type", None)
    description: str | None = kwargs.pop("description", None)
    branch = kwargs.pop("branch", None)
    branch_id = kwargs.pop("branch_id", 1)
    space = kwargs.pop("space", None)
    space_id = kwargs.pop("space_id", 1)
    _skip_validation = kwargs.pop("_skip_validation", False)
    if kwargs:
        valid_keywords = ", ".join([val[0] for val in _get_record_kwargs(Feature)])
        raise FieldValidationError(f"Only {valid_keywords} are valid keyword arguments")
    kwargs["name"] = name
    kwargs["type"] = type_
    kwargs["is_type"] = is_type
    kwargs["branch"] = branch
    kwargs["branch_id"] = branch_id
    kwargs["space"] = space
    kwargs["space_id"] = space_id
    kwargs["_skip_validation"] = _skip_validation
    if not is_param:
        kwargs["description"] = description
    # cast dtype
    if dtype is None and not is_type:
        raise ValidationError(
            f"Please pass dtype, one of {FEATURE_DTYPES} or a composed categorical dtype"
        )
    dtype_str = None
    if dtype is not None:
        if not isinstance(dtype, str):
            dtype_str = serialize_dtype(dtype)
        else:
            dtype_str = dtype
            parse_dtype(dtype_str, is_param=is_param)
        kwargs["dtype"] = dtype_str
    return kwargs


class Feature(SQLRecord, CanCurate, TracksRun, TracksUpdates):
    """Variables, such as dataframe columns or run parameters.

    A feature often represents a dimension of a dataset, such as a column in a `DataFrame`.
    The `Feature` registry organizes metadata of features.

    The `Feature` registry helps you organize and query datasets based on their
    features and corresponding label annotations. For instance, when working
    with a "T cell" label, it could be measured through different features
    such as `"cell_type_by_expert"` where an expert manually classified the
    cell, or `"cell_type_by_model"` where a computational model made the classification.

    The two most important metadata of a feature are its `name` and the `dtype`.
    In addition to typical data types, LaminDB has a `"num"` `dtype` to
    concisely denote the union of all numerical types.

    Args:
        name: `str` Name of the feature, typically a column name.
        dtype: `Dtype | Registry | list[Registry] | FieldAttr` See :class:`~lamindb.base.types.Dtype`.
            For categorical types, you can define to which registry values are
            restricted, e.g., `ULabel` or `[ULabel, bionty.CellType]`.
        unit: `str | None = None` Unit of measure, ideally SI (`"m"`, `"s"`, `"kg"`, etc.) or `"normalized"` etc.
        description: `str | None = None` A description.
        synonyms: `str | None = None` Bar-separated synonyms.
        nullable: `bool = True` Whether the feature can have null-like values (`None`, `pd.NA`, `NaN`, etc.), see :attr:`~lamindb.Feature.nullable`.
        default_value: `Any | None = None` Default value for the feature.
        coerce_dtype: `bool = False` When True, attempts to coerce values to the specified dtype
            during validation, see :attr:`~lamindb.Feature.coerce_dtype`.
        cat_filters: `dict[str, str] | None = None` Subset a registry by additional filters to define valid categories.

    Note:

        For more control, you can use :mod:`bionty` registries to manage simple
        biological entities like genes, proteins & cell markers. Or you define
        custom registries to manage high-level derived features like gene sets.

    See Also:
        :meth:`~lamindb.Feature.from_df`
            Create feature records from DataFrame.
        :attr:`~lamindb.Artifact.features`
            Feature manager of an artifact or collection.
        :class:`~lamindb.ULabel`
            Universal labels.
        :class:`~lamindb.Schema`
            Feature sets.

    Example:

        A simple `"str"` feature.::

            ln.Feature(name="sample_note", dtype=str).save()

        A dtype `"cat[ULabel]"` can be more easily passed as below.::

            ln.Feature(name="project", dtype=ln.ULabel).save()

        A dtype `"cat[ULabel|bionty.CellType]"` can be more easily passed as below.::

            ln.Feature(
                name="cell_type",
                dtype=[ln.ULabel, bt.CellType],
            ).save()

        A multivalue feature with a list of cell types.::

            ln.Feature(
                name="cell_types",
                dtype=list[bt.CellType],  # or list[str] for a list of strings
            ).save()

        A path feature.::

            ln.Feature(
                name="image_path",
                dtype="path",   # will be validated as `str`
            ).save()

    Hint:

        *Features* and *labels* denote two ways of using entities to organize data:

        1. A feature qualifies *what* is measured, i.e., a numerical or categorical random variable
        2. A label *is* a measured value, i.e., a category

        Consider annotating a dataset by that it measured expression of 30k
        genes: genes relate to the dataset as feature identifiers through a
        feature set with 30k members. Now consider annotating the artifact by
        whether that it measured the knock-out of 3 genes: here, the 3 genes act
        as labels of the dataset.

        Re-shaping data can introduce ambiguity among features & labels. If this
        happened, ask yourself what the joint measurement was: a feature
        qualifies variables in a joint measurement. The canonical data matrix
        lists jointly measured variables in the columns.
    """

    class Meta(SQLRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "lamindb"

    _name_field: str = "name"
    _aux_fields: dict[str, tuple[str, type]] = {
        "0": ("default_value", Any),  # type: ignore
        "1": ("nullable", bool),
        "2": ("coerce_dtype", bool),
    }

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(
        editable=False, unique=True, db_index=True, max_length=12, default=base62_12
    )
    """Universal id, valid across DB instances."""
    name: str = CharField(max_length=150, db_index=True)
    """Name of feature."""
    dtype: Dtype | None = CharField(db_index=True, null=True)
    """Data type (:class:`~lamindb.base.types.Dtype`)."""
    type: Feature | None = ForeignKey(
        "self", PROTECT, null=True, related_name="features"
    )
    """Type of feature (e.g., 'Readout', 'Metric', 'Metadata', 'ExpertAnnotation', 'ModelPrediction').

    Allows to group features by type, e.g., all read outs, all metrics, etc.
    """
    features: Feature
    """Features of this type (can only be non-empty if `is_type` is `True`)."""
    is_type: bool = BooleanField(default=False, db_index=True, null=True)
    """Distinguish types from instances of the type."""
    unit: str | None = CharField(max_length=30, db_index=True, null=True)
    """Unit of measure, ideally SI (`m`, `s`, `kg`, etc.) or 'normalized' etc. (optional)."""
    description: str | None = CharField(db_index=True, null=True)
    """A description."""
    array_rank: int = models.SmallIntegerField(default=0, db_index=True)
    """Rank of feature.

    Number of indices of the array: 0 for scalar, 1 for vector, 2 for matrix.

    Is called `.ndim` in `numpy` and `pytorch` but shouldn't be confused with
    the dimension of the feature space.
    """
    array_size: int = models.IntegerField(default=0, db_index=True)
    """Number of elements of the feature.

    Total number of elements (product of shape components) of the array.

    - A number or string (a scalar): 1
    - A 50-dimensional embedding: 50
    - A 25 x 25 image: 625
    """
    array_shape: list[int] | None = JSONField(default=None, db_default=None, null=True)
    """Shape of the feature.

    - A number or string (a scalar): [1]
    - A 50-dimensional embedding: [50]
    - A 25 x 25 image: [25, 25]

    Is stored as a list rather than a tuple because it's serialized as JSON.
    """
    proxy_dtype: Dtype | None = CharField(default=None, null=True)
    """Proxy data type.

    If the feature is an image it's often stored via a path to the image file. Hence, while the dtype might be
    image with a certain shape, the proxy dtype would be str.
    """
    synonyms: str | None = TextField(null=True)
    """Bar-separated (|) synonyms (optional)."""
    # we define the below ManyToMany on the feature model because it parallels
    # how other registries (like Gene, Protein, etc.) relate to Schema
    # it makes the API more consistent
    schemas: Schema = models.ManyToManyField(
        "Schema", through="SchemaFeature", related_name="features"
    )
    """Feature sets linked to this feature."""
    _expect_many: bool = models.BooleanField(default=None, db_default=None, null=True)
    """Indicates whether values for this feature are expected to occur a single or multiple times for an artifact (default `None`).

    - if it's `True` (default), the values come from an observation-level aggregation and a dtype of `datetime` on the observation-level means `set[datetime]` on the artifact-level
    - if it's `False` it's an artifact-level value and datetime means datetime; this is an edge case because an arbitrary artifact would always be a set of arbitrary measurements that would need to be aggregated ("one just happens to measure a single cell line in that artifact")
    """
    _curation: dict[str, Any] = JSONField(default=None, db_default=None, null=True)
    # backward fields
    values: FeatureValue
    """Values for this feature."""

    @overload
    def __init__(
        self,
        name: str,
        dtype: Dtype | Registry | list[Registry] | FieldAttr,
        type: Feature | None = None,
        is_type: bool = False,
        unit: str | None = None,
        description: str | None = None,
        synonyms: str | None = None,
        nullable: bool = True,
        default_value: str | None = None,
        coerce_dtype: bool = False,
        cat_filters: dict[str, str] | None = None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        if len(args) == len(self._meta.concrete_fields):
            super().__init__(*args, **kwargs)
            return None
        default_value = kwargs.pop("default_value", None)
        nullable = kwargs.pop("nullable", True)  # default value of nullable
        cat_filters = kwargs.pop("cat_filters", None)
        coerce_dtype = kwargs.pop("coerce_dtype", False)
        kwargs = process_init_feature_param(args, kwargs)
        super().__init__(*args, **kwargs)
        self.default_value = default_value
        self.nullable = nullable
        self.coerce_dtype = coerce_dtype
        dtype_str = kwargs.pop("dtype", None)
        if cat_filters:
            if "|" in dtype_str:
                raise ValidationError(
                    f"cat_filters are incompatible with union dtypes: '{dtype_str}'"
                )
            if "]]" in dtype_str:
                raise ValidationError(
                    f"cat_filters are incompatible with nested dtypes: '{dtype_str}'"
                )

            # Validate filter values and SQLRecord attributes
            for filter_key, filter_value in cat_filters.items():
                if not filter_value or (
                    isinstance(filter_value, str) and not filter_value.strip()
                ):
                    raise ValidationError(f"Empty value in filter {filter_key}")
                # Check SQLRecord attributes for relation lookups
                if isinstance(filter_value, SQLRecord) and "__" in filter_key:
                    field_name = filter_key.split("__", 1)[1]
                    if not hasattr(filter_value, field_name):
                        raise ValidationError(
                            f"SQLRecord {filter_value.__class__.__name__} has no attribute '{field_name}' in filter {filter_key}"
                        )

            # If a SQLRecord is passed, we access its uid to apply a standard filter
            cat_filters = {
                f"{key}__uid"
                if (
                    is_sqlrecord := isinstance(filter, SQLRecord)
                    and hasattr(filter, "uid")
                )
                else key: filter.uid if is_sqlrecord else filter
                for key, filter in cat_filters.items()
            }

            fill_in = ", ".join(
                f"{key}='{value}'" for (key, value) in cat_filters.items()
            )
            dtype_str = dtype_str.replace("]", f"[{fill_in}]]")
            self.dtype = dtype_str
        if not self._state.adding:
            if not (
                self.dtype.startswith("cat")
                if dtype_str == "cat"
                else dtype_str.startswith("cat")
                if self.dtype == "cat"
                else self.dtype == dtype_str
            ):
                raise ValidationError(
                    f"Feature {self.name} already exists with dtype {self.dtype}, you passed {dtype_str}"
                )

    @classmethod
    def from_dataframe(
        cls, df: pd.DataFrame, field: FieldAttr | None = None, *, mute: bool = False
    ) -> SQLRecordList:
        """Create Feature records for dataframe columns.

        Args:
            df: Source DataFrame to extract column information from
            field: FieldAttr for Feature model validation, defaults to Feature.name
            mute: Whether to mute Feature creation similar names found warnings
        """
        field = Feature.name if field is None else field
        registry = field.field.model  # type: ignore
        if registry != Feature:
            raise ValueError("field must be a Feature FieldAttr!")

        categoricals = categoricals_from_df(df)
        dtypes = {}
        for name, col in df.items():
            if name in categoricals:
                dtypes[name] = "cat"
            else:
                dtypes[name] = serialize_pandas_dtype(col.dtype)

        if mute:
            original_verbosity = logger._verbosity
            logger.set_verbosity(0)
        try:
            features = [
                Feature(name=name, dtype=dtype) for name, dtype in dtypes.items()
            ]  # type: ignore
            assert len(features) == len(df.columns)  # noqa: S101
            return SQLRecordList(features)
        finally:
            if mute:
                logger.set_verbosity(original_verbosity)

    @classmethod
    @deprecated("from_dataframe")
    def from_df(
        cls, df: pd.DataFrame, field: FieldAttr | None = None, *, mute: bool = False
    ) -> SQLRecordList:
        return cls.from_dataframe(df, field, mute=mute)

    @classmethod
    def from_dict(
        cls,
        dictionary: dict[str, Any],
        field: FieldAttr | None = None,
        *,
        str_as_cat: bool | None = None,
        mute: bool = False,
    ) -> SQLRecordList:
        """Create Feature records for dictionary keys.

        Args:
            dictionary: Source dictionary to extract key information from
            field: FieldAttr for Feature model validation, defaults to Feature.name
            str_as_cat: Whether to interpret string values as categorical
            mute: Whether to mute dtype inference and feature creation warnings
        """
        from lamindb.models._feature_manager import infer_feature_type_convert_json

        field = Feature.name if field is None else field
        registry = field.field.model  # type: ignore
        if registry != Feature:
            raise ValueError("field must be a Feature FieldAttr!")

        dtypes = {}
        ambiguous_keys = []
        for key, value in dictionary.items():
            dtype, _, message = infer_feature_type_convert_json(key, value, mute=mute)

            if dtype == "cat ? str":
                if str_as_cat is None:
                    ambiguous_keys.append(
                        (key, "str or cat", message.strip("# ") if message else "")
                    )
                    continue
                if str_as_cat:
                    dtype = "cat"
                else:
                    dtype = "str"

            elif dtype == "list[cat ? str]":
                if str_as_cat is None:
                    ambiguous_keys.append(
                        (
                            key,
                            "list[str] or list[cat]",
                            message.strip("# ") if message else "",
                        )
                    )
                    continue
                if str_as_cat:
                    dtype = "list[cat]"
                else:
                    dtype = "list[str]"

            dtypes[key] = dtype

        if ambiguous_keys:
            error_msg = "Ambiguous dtypes detected. Please pass `str_as_cat` parameter or create features explicitly:\n"
            for key, options, msg in ambiguous_keys:
                error_msg += f"  '{key}': {options}"
                if msg:
                    error_msg += f" ({msg})"
                error_msg += "\n"
            error_msg += "\nUse `str_as_cat=True` to treat strings as categorical, or `str_as_cat=False` for plain strings."
            raise ValueError(error_msg)

        if mute:
            original_verbosity = logger._verbosity
            logger.set_verbosity(0)
        try:
            features = [Feature(name=key, dtype=dtype) for key, dtype in dtypes.items()]  # type: ignore
            assert len(features) == len(dictionary)  # noqa: S101
            return SQLRecordList(features)
        finally:
            if mute:
                logger.set_verbosity(original_verbosity)

    def save(self, *args, **kwargs) -> Feature:
        """Save the feature to the instance."""
        super().save(*args, **kwargs)
        return self

    def with_config(self, optional: bool | None = None) -> tuple[Feature, dict]:
        """Pass addtional configurations to the schema."""
        if optional is not None:
            return self, {"optional": optional}
        return self, {}

    @property
    def default_value(self) -> Any:
        """A default value that overwrites missing values (default `None`).

        This takes effect when you call `Curator.standardize()`.

        If `default_value = None`, missing values like `pd.NA` or `np.nan` are kept.
        """
        if self._aux is not None and "af" in self._aux and "0" in self._aux["af"]:  # type: ignore
            return self._aux["af"]["0"]  # type: ignore
        else:
            return None

    @default_value.setter
    def default_value(self, value: str | None) -> None:
        self._aux = self._aux or {}
        self._aux.setdefault("af", {})["0"] = value

    @property
    def nullable(self) -> bool:
        """Indicates whether the feature can have nullable values (default `True`).

        Example::

            import lamindb as ln
            import pandas as pd

            disease = ln.Feature(name="disease", dtype=ln.ULabel, nullable=False).save()
            schema = ln.Schema(features=[disease]).save()
            dataset = {"disease": pd.Categorical([pd.NA, "asthma"])}
            df = pd.DataFrame(dataset)
            curator = ln.curators.DataFrameCurator(df, schema)
            try:
                curator.validate()
            except ln.errors.ValidationError as e:
                assert str(e).startswith("non-nullable series 'disease' contains null values")

        """
        if self._aux is not None and "af" in self._aux and "1" in self._aux["af"]:
            value = self._aux["af"]["1"]
            return True if value is None else value
        else:
            return True

    @nullable.setter
    def nullable(self, value: bool) -> None:
        assert isinstance(value, bool), value  # noqa: S101
        self._aux = self._aux or {}
        self._aux.setdefault("af", {})["1"] = value

    @property
    def coerce_dtype(self) -> bool:
        """Whether dtypes should be coerced during validation.

        For example, a `objects`-dtyped pandas column can be coerced to `categorical` and would pass validation if this is true.
        """
        if self._aux is not None and "af" in self._aux and "2" in self._aux["af"]:  # type: ignore
            return self._aux["af"]["2"]  # type: ignore
        else:
            return False

    @coerce_dtype.setter
    def coerce_dtype(self, value: bool) -> None:
        self._aux = self._aux or {}
        self._aux.setdefault("af", {})["2"] = value

    # we'll enable this later
    # @property
    # def observational_unit(self) -> Literal["Artifact", "Observation"]:
    #     """Default observational unit on which the feature is measured.

    #     Currently, we only make a distinction between artifact-level and observation-level features.

    #     For example, a feature `"ml_split"` that stores `"test"` & `"train"` labels is typically defined on the artifact level.
    #     When accessing `artifact.features.get_values(["ml_split"])`, you expect a single value, either `"test"` or `"train"`.

    #     However, when accessing an artifact annotation with a feature that's defined on the observation-level, say `"cell_type"`, you expect a set of values. So,
    #     `artifact.features.get_values(["cell_type_from_expert"])` should return a set: `{"T cell", "B cell"}`.

    #     The value of `observational_unit` is currently auto-managed: if using `artifact.features.add_values()`,
    #     it will be set to `Artifact`. In a curator, the value depends on whether it's an artifact- or observation-level slot
    #     (e.g. `.uns` is artifact-level in `AnnData` whereas `.obs` is observation-level).

    #     Note: This attribute might in the future be used to distinguish different types of observational units (e.g. single cells vs. physical samples vs. study subjects etc.).
    #     """
    #     if self._expect_many:
    #         return "Observation"  # this here might be replaced with the specific observational unit
    #     else:
    #         return "Artifact"


class FeatureValue(SQLRecord, TracksRun):
    """Non-categorical features values.

    Categorical feature values are stored in their respective registries:
    :class:`~lamindb.ULabel`, :class:`~bionty.CellType`, etc.

    Unlike for ULabel, in `FeatureValue`, values are grouped by features and
    not by an ontological hierarchy.
    """

    # we do not have a unique constraint on feature & value because it leads to hashing errors
    # for large dictionaries: https://lamin.ai/laminlabs/lamindata/transform/jgTrkoeuxAfs0000
    # we do not hash values because we have `get_or_create` logic all over the place
    # and also for checking whether the (feature, value) combination exists
    # there does not seem an issue with querying for a dict-like value
    # https://lamin.ai/laminlabs/lamindata/transform/jgTrkoeuxAfs0001

    _name_field: str = "value"

    feature: Feature | None = ForeignKey(
        Feature, CASCADE, null=True, related_name="values", default=None
    )
    """The dimension metadata."""
    value: Any = models.JSONField()
    """The JSON-like value."""
    hash: str = CharField(max_length=HASH_LENGTH, null=True, db_index=True)
    """Value hash."""

    class Meta(BaseSQLRecord.Meta, TracksRun.Meta):
        app_label = "lamindb"
        unique_together = ("feature", "hash")

    @classmethod
    def get_or_create(cls, feature, value):
        # simple values: (int, float, str, bool, datetime)
        if not isinstance(value, dict):
            hash = hash_string(str(value))
        else:
            hash = hash_dict(value)
        try:
            return (
                cls.objects.create(feature=feature, value=value, hash=hash),
                False,
            )
        except IntegrityError:
            return cls.objects.get(feature=feature, hash=hash), True


def suggest_categorical_for_str_iterable(
    iterable: Iterable[str], key: str = None
) -> str:
    c = pd.Categorical(iterable)
    message = ""
    if len(c.categories) < len(c):
        if key != "":
            key_note = f" for feature {key}"
        else:
            key_note = ""
        message = f"You have few permissible values{key_note}, consider dtype 'cat' instead of 'str'"
    return message


def categoricals_from_df(df: pd.DataFrame) -> dict:
    """Returns categorical columns."""
    string_cols = [col for col in df.columns if is_string_dtype(df[col])]
    categoricals = {
        col: df[col]
        for col in df.columns
        if isinstance(df[col].dtype, CategoricalDtype)
    }
    for key in string_cols:
        message = suggest_categorical_for_str_iterable(df[key], key)
        if message:
            logger.warning(message)
    return categoricals



import lamindb as ln


def test_delete_record():
    names = ["label1", "label2", "label3"]
    labels = [ln.ULabel(name=name) for name in names]
    ln.save(labels)
    ln.ULabel.filter(name__in=names).delete()
    assert ln.ULabel.filter(name__in=names).count() == 0



# ruff: noqa: S608
from django.db import migrations, models


def update_model_run_relationships(apps, schema_editor, model_name):
    vendor = schema_editor.connection.vendor

    # Define table names based on model_name
    model_table = f"lamindb_{model_name}"
    link_table = f"lamindb_{model_name}__previous_runs"
    model_id_field = f"{model_name}_id"

    with schema_editor.connection.cursor() as cursor:
        # Step 1: Add the current run_id to the _previous_runs table if it doesn't exist
        cursor.execute(f"""
            INSERT INTO {link_table} ({model_id_field}, run_id)
            SELECT a.id, a.run_id
            FROM {model_table} a
            WHERE a.run_id IS NOT NULL
            AND NOT EXISTS (
                SELECT 1
                FROM {link_table} apr
                WHERE apr.{model_id_field} = a.id
                AND apr.run_id = a.run_id
            );
        """)

        # Step 2: For each model, find the earliest run (lowest ID) and set it as the run_id
        if vendor == "sqlite":
            cursor.execute(f"""
                UPDATE {model_table}
                SET run_id = (
                    SELECT MIN(r.id)
                    FROM lamindb_run r
                    JOIN {link_table} apr ON r.id = apr.run_id
                    WHERE apr.{model_id_field} = {model_table}.id
                )
                WHERE EXISTS (
                    SELECT 1
                    FROM {link_table} apr
                    WHERE apr.{model_id_field} = {model_table}.id
                );
            """)
        else:  # PostgreSQL
            cursor.execute(f"""
                UPDATE {model_table} AS a
                SET run_id = subquery.min_run_id
                FROM (
                    SELECT {model_id_field}, MIN(run_id) as min_run_id
                    FROM {link_table}
                    GROUP BY {model_id_field}
                ) AS subquery
                WHERE a.id = subquery.{model_id_field}
                AND EXISTS (
                    SELECT 1
                    FROM {link_table} apr
                    WHERE apr.{model_id_field} = a.id
                );
            """)

        # Step 3: Remove the earliest run from the link table
        if vendor == "sqlite":
            cursor.execute(f"""
                DELETE FROM {link_table}
                WHERE EXISTS (
                    SELECT 1 FROM {model_table} a
                    WHERE {link_table}.{model_id_field} = a.id
                    AND {link_table}.run_id = a.run_id
                );
            """)
        else:  # PostgreSQL
            cursor.execute(f"""
                DELETE FROM {link_table} AS apr
                USING {model_table} AS a
                WHERE apr.{model_id_field} = a.id
                AND apr.run_id = a.run_id;
            """)


def update_artifact_run_relationships(apps, schema_editor):
    """Migration function for artifacts."""
    update_model_run_relationships(apps, schema_editor, "artifact")


def update_collection_run_relationships(apps, schema_editor):
    """Migration function for collections."""
    update_model_run_relationships(apps, schema_editor, "collection")


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0088_schema_components"),
    ]

    operations = [
        # unrelated to subsequent runs, but related to lamindb 1.2
        # update the otype field in the artifact table
        migrations.RunSQL(
            sql="""
            UPDATE lamindb_artifact
            SET otype = 'SpatialData'
            WHERE otype = 'spatialdata';
            """
        ),
        # Migrate artifact relationships
        migrations.RunPython(
            update_artifact_run_relationships, migrations.RunPython.noop
        ),
        # Update artifact model state
        migrations.SeparateDatabaseAndState(
            # Database operations (none, to keep tables intact)
            [],
            # State operations (to update Django model only)
            [
                # Remove the old field from model state
                migrations.RemoveField(
                    model_name="artifact",
                    name="_previous_runs",
                ),
                # Add the new field with the same underlying table
                migrations.AddField(
                    model_name="artifact",
                    name="_subsequent_runs",
                    field=models.ManyToManyField(
                        "lamindb.run",
                        related_name="_recreated_artifacts",
                        db_table="lamindb_artifact__previous_runs",  # Keep the original table name
                    ),
                ),
            ],
        ),
        # Migrate collection relationships
        migrations.RunPython(
            update_collection_run_relationships, migrations.RunPython.noop
        ),
        # Update collection model state
        migrations.SeparateDatabaseAndState(
            # Database operations (none, to keep tables intact)
            [],
            # State operations (to update Django model only)
            [
                # Remove the old field from model state
                migrations.RemoveField(
                    model_name="collection",
                    name="_previous_runs",
                ),
                # Add the new field with the same underlying table
                migrations.AddField(
                    model_name="collection",
                    name="_subsequent_runs",
                    field=models.ManyToManyField(
                        "lamindb.run",
                        related_name="_recreated_collections",
                        db_table="lamindb_collection__previous_runs",  # Keep the original table name
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 5.2 on 2025-01-16 01:29

import django.db.models.deletion
from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0078_lamindbv1_part6c"),
    ]

    operations = [
        migrations.AlterField(
            model_name="rundata",
            name="value_json",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="tidytabledata",
            name="value_json",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="tidytable",
            name="schema",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="_tidytables",
                to="lamindb.schema",
            ),
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

import lamindb_setup as ln_setup
from django.db.models import ManyToManyField
from lamindb_setup._connect_instance import (
    get_owner_name_from_identifier,
    load_instance_settings,
)
from lamindb_setup.core._settings_store import instance_settings_file

from lamindb.models.sqlrecord import IsLink

if TYPE_CHECKING:
    from lamindb.models.sqlrecord import Registry, SQLRecord


def get_schema_modules(instance: str | None) -> set[str]:
    if instance is None or instance == "default":
        schema_modules = set(ln_setup.settings.instance.modules)
        schema_modules.add("core")
        return schema_modules
    owner, name = get_owner_name_from_identifier(instance)
    settings_file = instance_settings_file(name, owner)
    if settings_file.exists():
        modules = set(load_instance_settings(settings_file).modules)
    else:
        cache_filepath = (
            ln_setup.settings.cache_dir / f"instance--{owner}--{name}--uid.txt"
        )
        if cache_filepath.exists():
            modules = set(cache_filepath.read_text().split("\n")[1].split(","))
        else:
            raise ValueError(f"Instance {instance} not found")
    shared_schema_modules = set(ln_setup.settings.instance.modules).intersection(
        modules
    )
    shared_schema_modules.add("core")
    return shared_schema_modules


# this function here should likely be renamed
# it maps the __get_name_with_module__() onto the actual model
def dict_module_name_to_model_name(
    registry: Registry, instance: str | None = None
) -> dict[str, Registry]:
    schema_modules = get_schema_modules(instance)
    d: dict = {
        i.related_model.__get_name_with_module__(): i.related_model
        for i in registry._meta.related_objects
        if i.related_name is not None
        and i.related_model.__get_module_name__() in schema_modules
    }
    d.update(
        {
            i.related_model.__get_name_with_module__(): i.related_model
            for i in registry._meta.many_to_many
            if i.name is not None
            and i.related_model.__get_module_name__() in schema_modules
        }
    )
    return d


def dict_related_model_to_related_name(
    registry: type[SQLRecord], links: bool = False, instance: str | None = None
) -> dict[str, str]:
    def include(model: SQLRecord):
        return not links != issubclass(model, IsLink)

    schema_modules = get_schema_modules(instance)

    related_objects = registry._meta.related_objects + registry._meta.many_to_many
    d: dict = {
        record.related_model.__get_name_with_module__(): (
            record.related_name
            if not isinstance(record, ManyToManyField)
            else record.name
        )
        for record in related_objects
        if (
            record.name is not None
            and include(record.related_model)
            and record.related_model.__get_module_name__() in schema_modules
        )
    }
    return d


def get_related_name(features_type: type[SQLRecord]) -> str:
    from lamindb.models.schema import Schema

    candidates = [
        field.related_name
        for field in Schema._meta.related_objects
        if field.related_model == features_type
    ]
    if not candidates:
        raise ValueError(
            f"Can't create feature sets from {features_type.__name__} because it's not"
            " related to it!\nYou need to create a link model between Schema and"
            " your SQLRecord in your custom module.\nTo do so, add a"
            " line:\n_feature_sets = models.ManyToMany(Schema,"
            " related_name='mythings')\n"
        )
    return candidates[0]



# ruff: noqa: F811

import lamindb as ln
import pytest
from _dataset_fixtures import (  # noqa
    get_mini_csv,
)
from lamindb.models.save import prepare_error_message, store_artifacts


def test_bulk_save_and_update():
    label_names = [f"ULabel {i} new" for i in range(3)]
    labels = [ln.ULabel(name=name) for name in label_names]
    # test bulk creation of new records
    ln.save(labels)
    assert len(ln.ULabel.filter(name__in=label_names).distinct().all()) == 3
    labels[0].name = "ULabel 0 updated"
    # test bulk update of existing records
    ln.save(labels)
    assert len(ln.ULabel.filter(name__in=label_names).distinct().all()) == 2
    assert ln.ULabel.get(name="ULabel 0 updated")


def test_prepare_error_message(get_mini_csv):
    artifact = ln.Artifact(get_mini_csv, description="test")
    exception = Exception("exception")

    error = prepare_error_message([], [artifact], exception)
    assert error.startswith(
        "The following entries have been successfully uploaded and committed to the database"
    )

    error = prepare_error_message([artifact], [], exception)
    assert error.startswith("No entries were uploaded or committed to the database")


def test_save_data_object(get_mini_csv):
    artifact = ln.Artifact(get_mini_csv, description="test")
    artifact.save()
    assert artifact.path.exists()
    artifact.delete(permanent=True, storage=True)


def test_store_artifacts_acid(get_mini_csv):
    artifact = ln.Artifact(get_mini_csv, description="test")
    artifact._clear_storagekey = "test.csv"
    # errors on check_and_attempt_clearing
    with pytest.raises(FileNotFoundError):
        artifact.save()

    with pytest.raises(RuntimeError) as error:
        store_artifacts([artifact], using_key=None)
    assert str(error.exconly()).startswith(
        "RuntimeError: The following entries have been successfully uploaded"
    )

    artifact.delete(permanent=True)


def test_save_parents():
    import bionty as bt

    records = bt.CellLine.from_values(["HEPG2", "HUVEC"])
    ln.save(records)
    assert bt.CellLine.get("4ea731nb").parents.to_dataframe().shape[0] == 1
    bt.CellLine.filter().delete()


def test_save_batch_size():
    label_names = [f"ULabel {i} batch_size" for i in range(3)]
    labels = [ln.ULabel(name=name) for name in label_names]
    # test bulk creation of new records with batch size
    ln.save(labels, batch_size=2)
    assert ln.ULabel.filter(name__in=label_names).distinct().count() == 3



# Generated by Django 5.2 on 2025-07-26 18:50

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0116_remove_artifact_unique_artifact_storage_key_hash_and_more"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="artifact",
            name="unique_artifact_storage_key_hash",
        ),
        migrations.AddConstraint(
            model_name="artifact",
            constraint=models.UniqueConstraint(
                condition=models.Q(("key__isnull", False)),
                fields=("storage", "key", "hash"),
                name="unique_artifact_storage_key_hash_not_null",
            ),
        ),
        migrations.AddConstraint(
            model_name="artifact",
            constraint=models.UniqueConstraint(
                condition=models.Q(("key__isnull", True)),
                fields=("storage", "hash"),
                name="unique_artifact_storage_hash_null_key",
            ),
        ),
    ]



import lamindb as ln

df = ln.core.datasets.mini_immuno.get_dataset1(otype="DataFrame")
df.attrs = {"temperature": 21.6, "experiment_id": "EXP001"}

study_metadata_schema = ln.Schema(
    features=[
        ln.Feature(name="temperature", dtype=float).save(),
        ln.Feature(name="experiment_id", dtype=str).save(),
    ],
).save()
schema = ln.Schema(
    features=[ln.Feature(name="perturbation", dtype="str").save()],
    slots={"attrs": study_metadata_schema},
    otype="DataFrame",
).save()

curator = ln.curators.DataFrameCurator(df, schema=schema)
curator.validate()
artifact = curator.save_artifact(key="examples/df_with_attrs.parquet")
artifact.describe()



import concurrent.futures
from pathlib import Path

import lamindb as ln
import pandas as pd
import pytest


@ln.tracked()
def process_chunk(chunk_id: int) -> str:
    # Create a simple DataFrame
    df = pd.DataFrame(
        {"id": range(chunk_id * 10, (chunk_id + 1) * 10), "value": range(10)}
    )
    env_file = Path("tmp_requirements.txt")
    env_file.write_text("1")
    ln.Artifact(env_file, description="requirements.txt").save()
    # Save it as an artifact
    key = f"chunk_{chunk_id}.parquet"
    artifact = ln.Artifact.from_dataframe(df, key=key).save()
    return artifact.key


def test_tracked_parallel():
    param_type = ln.Feature(name="Script[test_tracked.py]", is_type=True).save()
    ln.Feature(name="chunk_id", dtype="int", type=param_type).save()

    with pytest.raises(RuntimeError) as err:
        process_chunk(4)
    assert (
        err.exconly()
        == "RuntimeError: Please track the global run context before using @ln.tracked(): ln.track()"
    )

    # Ensure tracking is on
    ln.track()

    # Number of parallel executions
    n_parallel = 3

    # Use ThreadPoolExecutor for parallel execution
    with concurrent.futures.ThreadPoolExecutor(max_workers=n_parallel) as executor:
        # Submit all tasks
        futures = [executor.submit(process_chunk, i) for i in range(n_parallel)]
        # Get results as they complete
        chunk_keys = [
            future.result() for future in concurrent.futures.as_completed(futures)
        ]

    # Verify results
    # Each execution should have created its own artifact with unique run
    print(f"Created artifacts with keys: {chunk_keys}")
    artifacts = [ln.Artifact.get(key=key) for key in chunk_keys]
    env_artifacts = ln.Artifact.filter(description="requirements.txt").all()
    print(env_artifacts.to_dataframe())

    # Check that we got the expected number of artifacts
    assert len(artifacts) == n_parallel
    assert len(env_artifacts) == 1

    # Verify each artifact has its own unique run
    runs = [artifact.run for artifact in artifacts]
    run_ids = [run.id for run in runs]
    print(f"Run IDs: {run_ids}")
    assert len(set(run_ids)) == n_parallel  # all runs should be unique

    # Verify each run has the correct start and finish times
    for run in runs:
        print(f"Run details: {run}")
        assert run.started_at is not None
        assert run.finished_at is not None
        assert run.started_at < run.finished_at

    # Clean up test artifacts
    for artifact in artifacts:
        artifact.delete(permanent=True)
    env_artifacts[0].delete(permanent=True)

    ln.context._uid = None
    ln.context._run = None
    ln.context._transform = None
    ln.context._path = None


if __name__ == "__main__":
    test_tracked_parallel()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, overload

from django.db import models
from lamin_utils import logger
from lamin_utils._base62 import increment_base62
from lamindb_setup.core.upath import LocalPathClasses, UPath

from lamindb.base import ids
from lamindb.base.fields import (
    BooleanField,
    CharField,
)

if TYPE_CHECKING:  # noqa
    from lamindb.models.query_set import QuerySet


class IsVersioned(models.Model):
    """Base class for versioned models."""

    class Meta:
        abstract = True

    _len_stem_uid: int

    version: str | None = CharField(max_length=30, null=True, db_index=True)
    """Version (default `None`).

    Defines version of a family of records characterized by the same `stem_uid`.

    Consider using `semantic versioning <https://semver.org>`__
    with `Python versioning <https://peps.python.org/pep-0440/>`__.
    """
    is_latest: bool = BooleanField(default=True, db_index=True)
    """Boolean flag that indicates whether a record is the latest in its version family."""

    @overload
    def __init__(self): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        self._revises = kwargs.pop("revises") if "revises" in kwargs else None
        super().__init__(*args, **kwargs)

    @property
    def stem_uid(self) -> str:
        """Universal id characterizing the version family.

        The full uid of a record is obtained via concatenating the stem uid and version information::

            stem_uid = random_base62(n_char)  # a random base62 sequence of length 12 (transform) or 16 (artifact, collection)
            version_uid = "0000"  # an auto-incrementing 4-digit base62 number
            uid = f"{stem_uid}{version_uid}"  # concatenate the stem_uid & version_uid

        """
        return self.uid[: self._len_stem_uid]  # type: ignore

    @property
    def versions(self) -> QuerySet:
        """Lists all records of the same version family.

        >>> new_artifact = ln.Artifact(df2, revises=artifact).save()
        >>> new_artifact.versions()
        """
        return (
            self.__class__.using(self._state.db)
            .filter(uid__startswith=self.stem_uid)
            .order_by("-created_at")
        )

    def _add_to_version_family(self, revises: IsVersioned, version: str | None = None):
        """Add current record to a version family.

        Args:
            revises: a record that belongs to the version family.
            version: semantic version of the record.
        """
        old_uid = self.uid  # type: ignore
        new_uid, revises = create_uid(revises=revises, version=version)
        if self.__class__.__name__ == "Artifact" and self._key_is_virtual:
            old_path = self.path
            new_path = get_new_path_from_uid(
                old_path=old_path, old_uid=old_uid, new_uid=new_uid
            )
            new_path = UPath(old_path).rename(new_path)
            logger.success(f"updated path from {old_path} to {new_path}!")
        self.uid = new_uid
        self.version = version
        self.save()
        logger.success(f"updated uid from {old_uid} to {new_uid}!")


def bump_version(
    version: str,
    bump_type: str = "minor",
    behavior: Literal["prompt", "error", "ignore"] = "error",
) -> str:
    """Bumps the version number by major or minor depending on the bump_type flag.

    Args:
        version: The current version in "MAJOR" or "MAJOR.MINOR" format.
        bump_type: The type of version bump, either 'major' or 'minor'.

    Returns:
        The new version string.
    """
    try:
        # Split the version into major and minor parts if possible
        parts = version.split(".")
        major = int(parts[0])
        minor = int(parts[1]) if len(parts) > 1 else 0

        if bump_type == "major":
            # Bump the major version and reset the minor version
            new_version = f"{major + 1}"
        elif bump_type == "minor":
            # Bump the minor version
            new_version = f"{major}.{minor + 1}"
        else:
            raise ValueError("bump_type must be 'major' or 'minor'")

    except (ValueError, IndexError):
        if behavior == "prompt":
            new_version = input(
                f"The current version is '{version}' - please type the new version: "
            )
        elif behavior == "error":
            raise ValueError(
                "Cannot auto-increment non-integer castable version, please provide"
                " manually"
            ) from None
        else:
            logger.warning("could not auto-increment version, fix '?' manually")
            new_version = "?"
    return new_version


def set_version(version: str | None = None, previous_version: str | None = None):
    """(Auto-) set version.

    If `version` is `None`, returns the stored version.
    Otherwise sets the version to the passed version.

    Args:
        version: Version string.
        previous_version: Previous version string.
    """
    if version is None and previous_version is not None:
        version = bump_version(previous_version, bump_type="major")
    return version


def create_uid(
    *,
    version: str | None = None,
    n_full_id: int = 20,
    revises: IsVersioned | None = None,
) -> tuple[str, IsVersioned | None]:
    """This also updates revises in case it's not the latest version.

    This is why it returns revises.
    """
    if revises is not None:
        if not revises.is_latest:
            # need one more request
            revises = revises.__class__.objects.get(
                is_latest=True, uid__startswith=revises.stem_uid
            )
            logger.warning(
                f"didn't pass the latest version in `revises`, retrieved it: {revises}"
            )
        suid = revises.stem_uid
        vuid = increment_base62(revises.uid[-4:])  # type: ignore
    else:
        suid = ids.base62(n_full_id - 4)
        vuid = "0000"
    if version is not None:
        if not isinstance(version, str):
            raise ValueError(
                "`version` parameter must be `None` or `str`, e.g., '0.1', '1', '2', etc."
            )
        if revises is not None:
            if version == revises.version:
                raise ValueError(
                    f"Please increment the previous version: '{revises.version}'"
                )
    return suid + vuid, revises


def get_new_path_from_uid(old_path: UPath, old_uid: str, new_uid: str):
    if isinstance(old_path, LocalPathClasses):
        # for local path, the rename target must be full path
        new_path = old_path.as_posix().replace(old_uid, new_uid)
    else:
        # for cloud path, the rename target must be the last part of the path
        new_path = old_path.name.replace(old_uid, new_uid)
    return new_path


def process_revises(
    revises: IsVersioned | None,
    version: str | None,
    key: str | None,
    description: str | None,
    type: type[IsVersioned],
) -> tuple[str, str, str, str, IsVersioned | None]:
    if revises is not None and not isinstance(revises, type):
        raise TypeError(f"`revises` has to be of type `{type.__name__}`")
    uid, revises = create_uid(
        revises=revises, version=version, n_full_id=type._len_full_uid
    )
    if revises is not None:
        if description is None:
            description = revises.description
        if key is None:
            key = revises.key
    return uid, version, key, description, revises



# Generated by Django 5.2 on 2025-05-27 11:29

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0101_alter_artifact_hash_alter_feature_name_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="writelog",
            name="branch_code",
        ),
        migrations.RemoveField(
            model_name="writelog",
            name="space_uid",
        ),
        migrations.AddField(
            model_name="writelog",
            name="branch",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AddField(
            model_name="writelog",
            name="space",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="writelog",
            name="run_uid",
            field=models.CharField(default="0000000000000000", max_length=20),
        ),
        migrations.AlterField(
            model_name="writelog",
            name="record_uid",
            field=models.JSONField(db_index=True, default=0),
            preserve_default=False,
        ),
        migrations.AlterModelOptions(
            name="migrationstate",
            options={"base_manager_name": "objects"},
        ),
        migrations.AlterModelOptions(
            name="tablestate",
            options={"base_manager_name": "objects"},
        ),
        migrations.AlterField(
            model_name="writelog",
            name="migration_state",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="lamindb.migrationstate"
            ),
        ),
        migrations.AlterField(
            model_name="writelog",
            name="table",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="lamindb.tablestate"
            ),
        ),
    ]



"""Core library.

Settings & context:

.. autosummary::
   :toctree: .

   Settings
   subsettings
   Context

Artifact loaders:

.. autosummary::
   :toctree: .

   loaders

Data loaders:

.. autosummary::
   :toctree: .

   MappedCollection

Modules:

.. autosummary::
   :toctree: .

   storage
   logger

"""

from lamin_utils import logger
from lamin_utils._inspect import InspectResult

from .. import errors as exceptions
from ..examples import datasets  # backward compat
from . import loaders, subsettings, types
from ._context import Context
from ._mapped_collection import MappedCollection
from ._settings import Settings



import shutil
from subprocess import DEVNULL, run
from time import perf_counter

import lamindb_setup as ln_setup
import pytest
from lamin_utils import logger


def pytest_sessionstart():
    t_execute_start = perf_counter()
    # these are called in separate scripts because can't change connection
    # within the same python process due to django
    # init instance and setup RLS
    run(  # noqa: S602
        "python ./tests/permissions/scripts/setup_instance.py",
        shell=True,
        capture_output=False,
    )
    # populate permissions and models via the admin connection
    run(  # noqa: S602
        "python ./tests/permissions/scripts/setup_access.py",
        shell=True,
        capture_output=False,
    )

    total_time_elapsed = perf_counter() - t_execute_start
    print(f"time to setup the instance: {total_time_elapsed:.1f}s")


def pytest_sessionfinish(session: pytest.Session):
    logger.set_verbosity(1)
    shutil.rmtree("./default_storage_permissions")
    ln_setup.delete("lamindb-test-permissions", force=True)
    run("docker stop pgtest && docker rm pgtest", shell=True, stdout=DEVNULL)  # noqa: S602



# Generated by Django 5.2 on 2024-11-15 15:48

import django.db.models.deletion
from django.db import migrations, models

import lamindb.base.fields
import lamindb.base.ids
import lamindb.base.users
import lamindb.models


class Migration(migrations.Migration):
    initial = True
    # replaces = [
    #     (
    #         "lamindb",
    #         "0037_rename_dataset_to_collection",
    #     ),
    #     (
    #         "lamindb",
    #         "0038_alter_collection_artifact_alter_collection_artifacts_and_more",
    #     ),
    #     (
    #         "lamindb",
    #         "0039_remove_collection_artifacts_collectionartifact_and_more",
    #     ),
    #     ("lamindb", "0040_alter_transform_type"),
    #     (
    #         "lamindb",
    #         "0041_remove_transform_short_name_transform_description_and_more",
    #     ),
    #     ("lamindb", "0042_rename_run_at_run_started_at_run_finished_at"),
    #     ("lamindb", "0043_run_json"),
    #     ("lamindb", "0044_run_replicated_outputs_transform_ulabels_and_more"),
    #     (
    #         "lamindb",
    #         "0045_rename_replicated_outputs_run_replicated_output_artifacts_and_more",
    #     ),
    #     ("lamindb", "0046_storage_instance_uid"),
    #     ("lamindb", "0047_featurevalue_artifactfeaturevalue_and_more"),
    #     ("lamindb", "0048_alter_artifactulabel_feature_and_more"),
    #     ("lamindb", "0049_rename_type_feature_dtype_and_more"),
    #     ("lamindb", "0050_artifactfeatureset_feature_ref_is_semantic_and_more"),
    #     (
    #         "lamindb",
    #         "0051_remove_feature_feature_sets_featuresetfeature_and_more",
    #     ),
    #     (
    #         "lamindb",
    #         "0052_rename_feature_value_artifactfeaturevalue_featurevalue_and_more",
    #     ),
    #     ("lamindb", "0052_squashed"),
    #     (
    #         "lamindb",
    #         "0053_alter_featureset_hash_alter_paramvalue_created_by_and_more",
    #     ),
    #     ("lamindb", "0054_alter_feature_previous_runs_and_more"),
    #     ("lamindb", "0055_artifact_type_artifactparamvalue_and_more"),
    #     (
    #         "lamindb",
    #         "0056_rename_ulabel_ref_is_name_artifactulabel_label_ref_is_name_and_more",
    #     ),
    #     ("lamindb", "0057_link_models_latest_report_and_others"),
    #     ("lamindb", "0058_artifact__actions_collection__actions"),
    #     (
    #         "lamindb",
    #         "0059_alter_artifact__accessor_alter_artifact__hash_type_and_more",
    #     ),
    #     ("lamindb", "0060_alter_artifact__actions"),
    #     (
    #         "lamindb",
    #         "0061_alter_collection_meta_artifact_alter_run_environment_and_more",
    #     ),
    #     ("lamindb", "0062_add_is_latest_field"),
    #     ("lamindb", "0063_populate_latest_field"),
    #     (
    #         "lamindb",
    #         "0064_alter_artifact_version_alter_collection_version_and_more",
    #     ),
    #     ("lamindb", "0065_remove_collection_feature_sets_and_more"),
    #     ("lamindb", "0066_alter_artifact__feature_values_and_more"),
    #     ("lamindb", "0067_alter_featurevalue_unique_together_and_more"),
    #     ("lamindb", "0068_alter_artifactulabel_unique_together_and_more"),
    #     (
    #         "lamindb",
    #         "0069_alter_artifact__accessor_alter_artifact__hash_type_and_more",
    #     ),
    # ]

    dependencies = []  # type: ignore
    operations = [
        migrations.CreateModel(
            name="Feature",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(auto_now=True, db_index=True),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=lamindb.base.ids.base62_12,
                        max_length=12,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=150,
                        unique=True,
                    ),
                ),
                (
                    "dtype",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default=None, max_length=64
                    ),
                ),
                (
                    "unit",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "description",
                    lamindb.base.fields.TextField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "synonyms",
                    lamindb.base.fields.TextField(blank=True, default=None, null=True),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(lamindb.models.CanCurate, models.Model),
        ),
        migrations.CreateModel(
            name="FeatureSet",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=20,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=150, null=True
                    ),
                ),
                ("n", lamindb.base.fields.IntegerField(blank=True)),
                (
                    "dtype",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=64, null=True
                    ),
                ),
                (
                    "registry",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default=None, max_length=120
                    ),
                ),
                (
                    "hash",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=22,
                        null=True,
                        unique=True,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Param",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(auto_now=True, db_index=True),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default=None, max_length=100
                    ),
                ),
                (
                    "dtype",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default=None, max_length=64
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(lamindb.models.CanCurate, models.Model),
        ),
        migrations.CreateModel(
            name="User",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=8,
                        unique=True,
                    ),
                ),
                (
                    "handle",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=150,
                        null=True,
                    ),
                ),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(auto_now=True, db_index=True),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(models.Model, lamindb.models.CanCurate),
        ),
        migrations.CreateModel(
            name="Artifact",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(auto_now=True, db_index=True),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=20,
                        unique=True,
                    ),
                ),
                (
                    "description",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "key",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "suffix",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default=None, max_length=30
                    ),
                ),
                (
                    "type",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=20,
                        null=True,
                    ),
                ),
                (
                    "size",
                    lamindb.base.fields.BigIntegerField(
                        blank=True, db_index=True, null=True
                    ),
                ),
                (
                    "hash",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=22,
                        null=True,
                    ),
                ),
                (
                    "n_objects",
                    lamindb.base.fields.BigIntegerField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "n_observations",
                    lamindb.base.fields.BigIntegerField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "_hash_type",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "_accessor",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=64,
                        null=True,
                    ),
                ),
                (
                    "visibility",
                    models.SmallIntegerField(
                        choices=[(1, "Default"), (0, "Hidden"), (-1, "Trash")],
                        db_index=True,
                        default=1,
                    ),
                ),
                (
                    "_key_is_virtual",
                    lamindb.base.fields.BooleanField(blank=True, default=None),
                ),
                (
                    "_actions",
                    models.ManyToManyField(
                        related_name="_action_targets", to="lamindb.artifact"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Collection",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(auto_now=True, db_index=True),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=lamindb.base.ids.base62_20,
                        max_length=20,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default=None, max_length=150
                    ),
                ),
                (
                    "description",
                    lamindb.base.fields.TextField(blank=True, default=None, null=True),
                ),
                (
                    "hash",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=22,
                        null=True,
                    ),
                ),
                (
                    "reference",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "reference_type",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=25,
                        null=True,
                    ),
                ),
                (
                    "visibility",
                    models.SmallIntegerField(
                        choices=[(1, "Default"), (0, "Hidden"), (-1, "Trash")],
                        db_index=True,
                        default=1,
                    ),
                ),
                (
                    "_actions",
                    models.ManyToManyField(related_name="+", to="lamindb.artifact"),
                ),
                (
                    "meta_artifact",
                    lamindb.base.fields.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="_meta_of_collection",
                        to="lamindb.artifact",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="CollectionArtifact",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "artifact",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_collection",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "collection",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_artifact",
                        to="lamindb.collection",
                    ),
                ),
            ],
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="collection",
            name="artifacts",
            field=models.ManyToManyField(
                related_name="collections",
                through="lamindb.CollectionArtifact",
                to="lamindb.artifact",
            ),
        ),
        migrations.CreateModel(
            name="ArtifactFeatureSet",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "slot",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=40, null=True
                    ),
                ),
                (
                    "feature_ref_is_semantic",
                    lamindb.base.fields.BooleanField(
                        blank=True, default=None, null=True
                    ),
                ),
                (
                    "artifact",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_feature_set",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "featureset",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_artifact",
                        to="lamindb.featureset",
                    ),
                ),
            ],
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="artifact",
            name="feature_sets",
            field=models.ManyToManyField(
                related_name="artifacts",
                through="lamindb.ArtifactFeatureSet",
                to="lamindb.featureset",
            ),
        ),
        migrations.CreateModel(
            name="FeatureSetFeature",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "featureset",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.featureset",
                    ),
                ),
            ],
            options={
                "unique_together": {("featureset", "feature")},
            },
            bases=(models.Model, lamindb.models.IsLink),
        ),
        migrations.AddField(
            model_name="feature",
            name="feature_sets",
            field=models.ManyToManyField(
                related_name="features",
                through="lamindb.FeatureSetFeature",
                to="lamindb.featureset",
            ),
        ),
        migrations.CreateModel(
            name="FeatureValue",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("value", models.JSONField()),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values",
                        to="lamindb.feature",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ArtifactFeatureValue",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "artifact",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "featurevalue",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.featurevalue",
                    ),
                ),
            ],
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="artifact",
            name="_feature_values",
            field=models.ManyToManyField(
                related_name="artifacts",
                through="lamindb.ArtifactFeatureValue",
                to="lamindb.featurevalue",
            ),
        ),
        migrations.CreateModel(
            name="ParamValue",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.JSONField()),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "param",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values",
                        to="lamindb.param",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ArtifactParamValue",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "artifact",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "paramvalue",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.paramvalue",
                    ),
                ),
            ],
            options={
                "unique_together": {("artifact", "paramvalue")},
            },
            bases=(models.Model, lamindb.models.IsLink),
        ),
        migrations.AddField(
            model_name="artifact",
            name="_param_values",
            field=models.ManyToManyField(
                related_name="artifacts",
                through="lamindb.ArtifactParamValue",
                to="lamindb.paramvalue",
            ),
        ),
        migrations.CreateModel(
            name="Run",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=lamindb.base.ids.base62_20,
                        max_length=20,
                        unique=True,
                    ),
                ),
                (
                    "started_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "finished_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "is_consecutive",
                    lamindb.base.fields.BooleanField(
                        blank=True, default=None, null=True
                    ),
                ),
                (
                    "reference",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "reference_type",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=25,
                        null=True,
                    ),
                ),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "environment",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="_environment_of",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "parent",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="children",
                        to="lamindb.run",
                    ),
                ),
                (
                    "report",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="_report_of",
                        to="lamindb.artifact",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="param",
            name="_previous_runs",
            field=models.ManyToManyField(related_name="+", to="lamindb.run"),
        ),
        migrations.AddField(
            model_name="param",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.models.current_run,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.run",
            ),
        ),
        migrations.AddField(
            model_name="featurevalue",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.models.current_run,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.run",
            ),
        ),
        migrations.AddField(
            model_name="featureset",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.models.current_run,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.run",
            ),
        ),
        migrations.AddField(
            model_name="feature",
            name="_previous_runs",
            field=models.ManyToManyField(related_name="+", to="lamindb.run"),
        ),
        migrations.AddField(
            model_name="feature",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.models.current_run,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.run",
            ),
        ),
        migrations.AddField(
            model_name="collectionartifact",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.models.current_run,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.run",
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="_previous_runs",
            field=models.ManyToManyField(
                related_name="_output_collections_with_later_updates",
                to="lamindb.run",
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="input_of_runs",
            field=models.ManyToManyField(
                related_name="input_collections", to="lamindb.run"
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="output_collections",
                to="lamindb.run",
            ),
        ),
        migrations.AddField(
            model_name="artifactfeaturevalue",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.models.current_run,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.run",
            ),
        ),
        migrations.AddField(
            model_name="artifactfeatureset",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.models.current_run,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.run",
            ),
        ),
        migrations.AddField(
            model_name="artifact",
            name="_previous_runs",
            field=models.ManyToManyField(
                related_name="_output_artifacts_with_later_updates",
                to="lamindb.run",
            ),
        ),
        migrations.AddField(
            model_name="artifact",
            name="input_of_runs",
            field=models.ManyToManyField(
                related_name="input_artifacts", to="lamindb.run"
            ),
        ),
        migrations.AddField(
            model_name="artifact",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="output_artifacts",
                to="lamindb.run",
            ),
        ),
        migrations.CreateModel(
            name="RunParamValue",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "paramvalue",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.paramvalue",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("run", "paramvalue")},
            },
            bases=(models.Model, lamindb.models.IsLink),
        ),
        migrations.AddField(
            model_name="run",
            name="_param_values",
            field=models.ManyToManyField(
                related_name="runs",
                through="lamindb.RunParamValue",
                to="lamindb.paramvalue",
            ),
        ),
        migrations.CreateModel(
            name="Storage",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(auto_now=True, db_index=True),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=lamindb.base.ids.base62_12,
                        max_length=12,
                        unique=True,
                    ),
                ),
                (
                    "root",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=255,
                        unique=True,
                    ),
                ),
                (
                    "description",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "type",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default=None, max_length=30
                    ),
                ),
                (
                    "region",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=64,
                        null=True,
                    ),
                ),
                (
                    "instance_uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=12,
                        null=True,
                    ),
                ),
                (
                    "_previous_runs",
                    models.ManyToManyField(related_name="+", to="lamindb.run"),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="artifact",
            name="storage",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifacts",
                to="lamindb.storage",
            ),
        ),
        migrations.CreateModel(
            name="Transform",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=16,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=150,
                        null=True,
                    ),
                ),
                (
                    "key",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=120,
                        null=True,
                    ),
                ),
                (
                    "description",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=255, null=True
                    ),
                ),
                (
                    "type",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default="pipeline", max_length=20
                    ),
                ),
                (
                    "source_code",
                    lamindb.base.fields.TextField(blank=True, default=None, null=True),
                ),
                (
                    "hash",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=22,
                        null=True,
                    ),
                ),
                (
                    "reference",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "reference_type",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=25,
                        null=True,
                    ),
                ),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(auto_now=True, db_index=True),
                ),
                (
                    "_source_code_artifact",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="_source_code_of",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "predecessors",
                    models.ManyToManyField(
                        related_name="successors", to="lamindb.transform"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="run",
            name="transform",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="runs",
                to="lamindb.transform",
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="transform",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="output_collections",
                to="lamindb.transform",
            ),
        ),
        migrations.AddField(
            model_name="artifact",
            name="transform",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="output_artifacts",
                to="lamindb.transform",
            ),
        ),
        migrations.CreateModel(
            name="ULabel",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(auto_now=True, db_index=True),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=lamindb.base.ids.base62_8,
                        max_length=8,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=150,
                        unique=True,
                    ),
                ),
                (
                    "description",
                    lamindb.base.fields.TextField(blank=True, default=None, null=True),
                ),
                (
                    "reference",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "reference_type",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=25,
                        null=True,
                    ),
                ),
                (
                    "_previous_runs",
                    models.ManyToManyField(related_name="+", to="lamindb.run"),
                ),
                (
                    "parents",
                    models.ManyToManyField(
                        related_name="children", to="lamindb.ulabel"
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(
                lamindb.models.HasParents,
                lamindb.models.CanCurate,
                models.Model,
            ),
        ),
        migrations.AddField(
            model_name="transform",
            name="ulabels",
            field=models.ManyToManyField(
                related_name="transforms", to="lamindb.ulabel"
            ),
        ),
        migrations.CreateModel(
            name="CollectionULabel",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "label_ref_is_name",
                    lamindb.base.fields.BooleanField(
                        blank=True, default=None, null=True
                    ),
                ),
                (
                    "feature_ref_is_name",
                    lamindb.base.fields.BooleanField(
                        blank=True, default=None, null=True
                    ),
                ),
                (
                    "collection",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_ulabel",
                        to="lamindb.collection",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_collectionulabel",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "ulabel",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_collection",
                        to="lamindb.ulabel",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
            ],
            options={
                "unique_together": {("collection", "ulabel")},
            },
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="collection",
            name="ulabels",
            field=models.ManyToManyField(
                related_name="collections",
                through="lamindb.CollectionULabel",
                to="lamindb.ulabel",
            ),
        ),
        migrations.CreateModel(
            name="ArtifactULabel",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "label_ref_is_name",
                    lamindb.base.fields.BooleanField(
                        blank=True, default=None, null=True
                    ),
                ),
                (
                    "feature_ref_is_name",
                    lamindb.base.fields.BooleanField(
                        blank=True, default=None, null=True
                    ),
                ),
                (
                    "artifact",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_ulabel",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_artifactulabel",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "ulabel",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_artifact",
                        to="lamindb.ulabel",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
            ],
            options={
                "unique_together": {("artifact", "ulabel", "feature")},
            },
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="artifact",
            name="ulabels",
            field=models.ManyToManyField(
                related_name="artifacts",
                through="lamindb.ArtifactULabel",
                to="lamindb.ulabel",
            ),
        ),
        migrations.AddField(
            model_name="transform",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="created_transforms",
                to="lamindb.user",
            ),
        ),
        migrations.AddField(
            model_name="run",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="created_runs",
                to="lamindb.user",
            ),
        ),
        migrations.AddField(
            model_name="param",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AddField(
            model_name="featurevalue",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AddField(
            model_name="featureset",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AddField(
            model_name="feature",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AddField(
            model_name="collectionartifact",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AddField(
            model_name="artifactfeaturevalue",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AddField(
            model_name="artifactfeatureset",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AddField(
            model_name="artifact",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="created_artifacts",
                to="lamindb.user",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="collectionartifact",
            unique_together={("collection", "artifact")},
        ),
        migrations.AlterUniqueTogether(
            name="artifactfeaturevalue",
            unique_together={("artifact", "featurevalue")},
        ),
        migrations.AlterUniqueTogether(
            name="artifactfeatureset",
            unique_together={("artifact", "featureset")},
        ),
    ]



# Generated by Django 5.2 on 2025-01-11 16:44

import django.db.models.deletion
from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0071_lamindbv1_migrate_schema"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="user",
            name="_branch_code",
        ),
        migrations.RemoveField(
            model_name="user",
            name="aux",
        ),
        migrations.RemoveField(
            model_name="user",
            name="space",
        ),
        migrations.AlterField(
            model_name="artifact",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="featureset",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="featurevalue",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="param",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="paramvalue",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="run",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="storage",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="transform",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
    ]



from pathlib import Path

import anndata as ad
import lamindb as ln
import pandas as pd
import pytest

# ruff: noqa: F811
from _dataset_fixtures import get_small_mdata, get_small_sdata  # noqa


@pytest.fixture(scope="module")
def zip_file():
    filepath = Path("test.zip")
    with open(filepath, "w") as f:
        f.write("some")
    yield filepath
    filepath.unlink()


@pytest.fixture(scope="module")
def html_filepath():
    filepath = Path("./tmp.html")
    with open(filepath, "w") as f:
        f.write("<html><body><h1>Test</h1></body></html>")
    yield filepath
    filepath.unlink()


@pytest.fixture(scope="module")
def json_filepath():
    filepath = Path("./tmp.json")
    with open(filepath, "w") as f:
        f.write('{"a": 1}')
    yield filepath
    filepath.unlink()


@pytest.fixture(scope="module")
def csv_filepath():
    filepath = Path("./tmp.csv")
    with open(filepath, "w") as f:
        f.write("a,b\n1,2")
    yield filepath
    filepath.unlink()


@pytest.fixture(scope="module")
def tsv_filepath():
    filepath = Path("./tmp.tsv")
    with open(filepath, "w") as f:
        f.write("a\tb\n1\t2")
    yield filepath
    filepath.unlink()


@pytest.fixture(scope="module")
def parquet_filepath():
    filepath = Path("./tmp.parquet")
    df = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    df.to_parquet(filepath)
    yield filepath
    filepath.unlink()


@pytest.fixture(scope="module")
def yaml_filepath():
    filepath = Path("./tmp.yaml")
    with open(filepath, "w") as f:
        f.write("a: 1\nb: 2")
    yield filepath
    filepath.unlink()


@pytest.fixture(scope="module")
def image_filepath():
    filepath = Path("./tmp.png")
    with open(filepath, "w") as f:
        f.write("mock image")
    yield filepath
    filepath.unlink()


@pytest.fixture(scope="module")
def svg_filepath():
    filepath = Path("./tmp.svg")
    with open(filepath, "w") as f:
        f.write("<svg><rect width='100' height='100'/></svg>")
    yield filepath
    filepath.unlink()


@pytest.fixture(scope="module")
def rds_filepath():
    filepath = Path("./tmp.rds")
    with open(filepath, "w") as f:
        f.write("mock rds")
    yield filepath
    filepath.unlink()


@pytest.fixture(scope="module")
def local_anndata_filepath():
    return ln.core.datasets.anndata_file_pbmc68k_test().resolve()


@pytest.fixture(scope="module")
def adata(local_anndata_filepath):
    return ad.read_h5ad(local_anndata_filepath)


def test_load_anndata(local_anndata_filepath, adata):
    artifact = ln.Artifact(local_anndata_filepath, description="test")
    assert local_anndata_filepath == artifact._local_filepath
    assert local_anndata_filepath == artifact.path
    assert local_anndata_filepath == artifact.cache()

    artifact = ln.Artifact.from_anndata(adata, description="test")
    assert artifact._memory_rep is adata
    assert artifact.load() is adata
    assert artifact._local_filepath.resolve() == artifact.cache() == artifact.path


def test_load_mudata(get_small_mdata):
    artifact = ln.Artifact.from_mudata(get_small_mdata, description="test")
    assert artifact._memory_rep is get_small_mdata
    assert artifact.load() is get_small_mdata
    assert artifact._local_filepath.resolve() == artifact.cache() == artifact.path


def test_load_spatialdata(get_small_sdata):
    artifact = ln.Artifact.from_spatialdata(get_small_sdata, description="test")
    assert artifact._memory_rep is get_small_sdata
    assert artifact.load() is get_small_sdata
    assert artifact._local_filepath.resolve() == artifact.cache() == artifact.path


def load_blobs__repr__():
    example_blobs_sdata = ln.core.datasets.spatialdata_blobs()
    blobs_af = ln.Artifact.from_spatialdata(
        example_blobs_sdata, key="example_blobs.zarr"
    ).save()
    example_blobs_sdata = blobs_af.load()
    # Must exist and not throw errors
    assert example_blobs_sdata.__repr__


def test_load_html(html_filepath):
    artifact = ln.Artifact(html_filepath, key=str(html_filepath))
    artifact.load()


def test_load_json(json_filepath):
    artifact = ln.Artifact(json_filepath, key=str(json_filepath))
    dictionary = artifact.load()
    assert dictionary["a"] == 1


def test_no_loader(zip_file):
    artifact = ln.Artifact(zip_file, key=str(zip_file))
    with pytest.raises(NotImplementedError):
        artifact.load()


def test_load_csv(csv_filepath):
    artifact = ln.Artifact(csv_filepath, key=str(csv_filepath))
    df = artifact.load()
    assert df.iloc[0, 0] == 1
    assert df.iloc[0, 1] == 2


def test_load_tsv(tsv_filepath):
    artifact = ln.Artifact(tsv_filepath, key=str(tsv_filepath))
    df = artifact.load()
    assert df.iloc[0, 0] == 1
    assert df.iloc[0, 1] == 2


def test_load_parquet(parquet_filepath):
    artifact = ln.Artifact(parquet_filepath, key=str(parquet_filepath))
    df = artifact.load()
    assert df.iloc[0, 0] == 1
    assert df.iloc[1, 1] == 4


def test_load_yaml(yaml_filepath):
    artifact = ln.Artifact(yaml_filepath, key=str(yaml_filepath))
    data = artifact.load()
    assert data["a"] == 1
    assert data["b"] == 2


def test_load_image(image_filepath):
    artifact = ln.Artifact(image_filepath, key=str(image_filepath))
    result = artifact.load()
    assert Path(result).name == image_filepath.name


def test_load_svg(svg_filepath):
    artifact = ln.Artifact(svg_filepath, key=str(svg_filepath))
    result = artifact.load()
    assert Path(result).name == svg_filepath.name


def test_load_rds(rds_filepath, ccaplog):
    artifact = ln.Artifact(rds_filepath, key=str(rds_filepath))
    result = artifact.load()
    assert "Please use `laminr` to load `.rds` files" in ccaplog.text
    assert Path(result).name == rds_filepath.name



import lamindb as ln
import bionty as bt

# define valid labels
perturbation_type = ln.ULabel(name="Perturbation", is_type=True).save()
ln.ULabel(name="DMSO", type=perturbation_type).save()
ln.ULabel(name="IFNG", type=perturbation_type).save()
bt.CellType.from_source(name="B cell").save()
bt.CellType.from_source(name="T cell").save()

# define valid features
ln.Feature(name="perturbation", dtype=perturbation_type).save()
ln.Feature(name="cell_type_by_expert", dtype=bt.CellType).save()
ln.Feature(name="cell_type_by_model", dtype=bt.CellType).save()
ln.Feature(name="assay_oid", dtype=bt.ExperimentalFactor.ontology_id).save()
ln.Feature(name="concentration", dtype=str).save()
ln.Feature(name="treatment_time_h", dtype="num", coerce_dtype=True).save()
ln.Feature(name="donor", dtype=str, nullable=True).save()
ln.Feature(name="donor_ethnicity", dtype=list[bt.Ethnicity]).save()



import json

import psycopg2


def sign_jwt(db_url, payload: dict) -> str:
    with psycopg2.connect(db_url) as conn, conn.cursor() as cur:
        cur.execute(
            """
                SELECT sign(
                    %s::json,
                    (SELECT security.get_secret('jwt_secret')),
                    %s
                )
                """,
            (json.dumps(payload), "HS256"),
        )
        token = cur.fetchone()[0]
        if not token:
            msg = "Failed to generate JWT"
            raise ValueError(msg)
        return token



# Generated by Django 5.2 on 2025-01-13 14:10

import django.db.models.deletion
from django.db import migrations, models

import lamindb.base.fields
import lamindb.base.users
import lamindb.models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0074_lamindbv1_part4"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="FeatureSet",
            new_name="Schema",
        ),
        migrations.RenameModel(
            old_name="ArtifactFeatureSet",
            new_name="ArtifactSchema",
        ),
        migrations.RenameField(
            model_name="feature",
            old_name="feature_sets",
            new_name="schemas",
        ),
        migrations.RenameField(
            model_name="artifact",
            old_name="feature_sets",
            new_name="_schemas_m2m",
        ),
        migrations.RenameField(
            model_name="artifactschema",
            old_name="featureset",
            new_name="schema",
        ),
        migrations.RenameModel(
            old_name="FeatureSetFeature",
            new_name="SchemaFeature",
        ),
        migrations.RenameField(
            model_name="schemafeature",
            old_name="featureset",
            new_name="schema",
        ),
        migrations.AddField(
            model_name="artifact",
            name="schema",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifacts",
                to="lamindb.schema",
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="_schemas_m2m",
            field=models.ManyToManyField(
                related_name="_artifacts_m2m",
                through="lamindb.ArtifactSchema",
                to="lamindb.schema",
            ),
        ),
        migrations.AlterField(
            model_name="artifactschema",
            name="artifact",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="_links_schema",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactschema",
            name="schema",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="_links_artifact",
                to="lamindb.schema",
            ),
        ),
        migrations.RemoveField(
            model_name="artifact",
            name="_curator",
        ),
        migrations.AddField(
            model_name="feature",
            name="_curation",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="feature",
            name="array_size",
            field=models.IntegerField(db_index=True, default=0),
        ),
        migrations.AddField(
            model_name="feature",
            name="proxy_dtype",
            field=lamindb.base.fields.CharField(
                blank=True, null=True, default=None, max_length=255
            ),
        ),
        migrations.AddField(
            model_name="feature",
            name="array_rank",
            field=models.SmallIntegerField(db_index=True, default=0),
        ),
        migrations.AddField(
            model_name="feature",
            name="array_shape",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="project",
            name="_status_code",
            field=models.SmallIntegerField(db_index=True, default=0),
        ),
        migrations.AddField(
            model_name="project",
            name="end_date",
            field=lamindb.base.fields.DateField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name="project",
            name="start_date",
            field=lamindb.base.fields.DateField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name="schema",
            name="_curation",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="schema",
            name="composite",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="components",
                to="lamindb.schema",
            ),
        ),
        migrations.AddField(
            model_name="schema",
            name="maximal_set",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False
            ),
        ),
        migrations.AddField(
            model_name="schema",
            name="minimal_set",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=True
            ),
        ),
        migrations.AddField(
            model_name="schema",
            name="ordered_set",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False
            ),
        ),
        migrations.AddField(
            model_name="schema",
            name="otype",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=64, null=True
            ),
        ),
        migrations.AddField(
            model_name="schema",
            name="slot",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=100, null=True
            ),
        ),
        migrations.AddField(
            model_name="schema",
            name="validated_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="validated_schemas",
                to="lamindb.schema",
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="dtype",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=22, null=True
            ),
        ),
        migrations.CreateModel(
            name="SchemaParam",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "param",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.param",
                    ),
                ),
                (
                    "schema",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.schema",
                    ),
                ),
            ],
            options={
                "unique_together": {("schema", "param")},
            },
            bases=(models.Model, lamindb.models.IsLink),
        ),
        migrations.AddField(
            model_name="param",
            name="schemas",
            field=models.ManyToManyField(
                related_name="params",
                through="lamindb.SchemaParam",
                to="lamindb.schema",
            ),
        ),
        migrations.RenameField(
            model_name="schema",
            old_name="registry",
            new_name="itype",
        ),
        migrations.AlterField(
            model_name="schema",
            name="itype",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=120, null=True
            ),
        ),
    ]


# if "bionty" in lamindb_setup.settings.instance.modules:
#     Migration.dependencies += [
#         ("bionty", "0046_alter_cellline__aux_alter_cellmarker__aux_and_more"),
#     ]



# Generated by Django 5.2 on 2025-08-07 18:52

from django.db import migrations

CREATE_FUNCTION_SQL = """
CREATE OR REPLACE FUNCTION is_valid_record_type(record_type_id INTEGER, record_is_type BOOLEAN)
RETURNS BOOLEAN AS $$
BEGIN
    -- Record with no type is valid
    IF record_type_id IS NULL THEN
        RETURN TRUE;
    END IF;

    -- If current record is a type, it can only reference schema-less types
    IF record_is_type THEN
        RETURN EXISTS (
            SELECT 1 FROM lamindb_record r
            WHERE r.id = record_type_id AND r.is_type AND r.schema_id IS NULL
        );
    END IF;

    -- Regular records can reference any type
    RETURN EXISTS (
        SELECT 1 FROM lamindb_record r
        WHERE r.id = record_type_id AND r.is_type
    );
END;
$$ LANGUAGE plpgsql;
"""

ADD_CONSTRAINT_SQL = """
ALTER TABLE lamindb_record
ADD CONSTRAINT record_type_is_valid_fk
CHECK (is_valid_record_type(type_id, is_type));
"""

DROP_CONSTRAINT_SQL = (
    "ALTER TABLE lamindb_record DROP CONSTRAINT IF EXISTS record_type_is_valid_fk;"
)
DROP_FUNCTION_SQL = "DROP FUNCTION IF EXISTS is_valid_record_type(INTEGER, BOOLEAN);"


def apply_postgres_constraint(apps, schema_editor):
    if schema_editor.connection.vendor == "postgresql":
        schema_editor.execute(CREATE_FUNCTION_SQL)
        schema_editor.execute(ADD_CONSTRAINT_SQL)


def revert_postgres_constraint(apps, schema_editor):
    if schema_editor.connection.vendor == "postgresql":
        schema_editor.execute(DROP_CONSTRAINT_SQL)
        schema_editor.execute(DROP_FUNCTION_SQL)


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0119_squashed"),
    ]

    operations = [
        migrations.RunPython(
            apply_postgres_constraint, reverse_code=revert_postgres_constraint
        ),
    ]



import lamindb as ln
from lamindb.models.has_parents import _add_emoji


def test_view_parents():
    label1 = ln.ULabel(name="label1")
    label2 = ln.ULabel(name="label2")
    label1.save()
    label2.save()
    label1.parents.add(label2)
    label1.view_parents(ln.ULabel.name, distance=1)
    label1.delete()
    label2.delete()


def test_query_parents_children():
    label1 = ln.ULabel(name="label1").save()
    label2 = ln.ULabel(name="label2").save()
    label3 = ln.ULabel(name="label3").save()
    label1.children.add(label2)
    label2.children.add(label3)
    parents = label3.query_parents()
    assert len(parents) == 2
    assert label1 in parents and label2 in parents
    children = label1.query_children()
    assert len(children) == 2
    assert label2 in children and label3 in children
    label1.delete(permanent=True)
    label2.delete(permanent=True)
    label3.delete(permanent=True)


def test_add_emoji():
    transform = ln.Transform(key="test-12345", type="upload")
    assert _add_emoji(transform, label="transform") == "🖥️ transform"
    transform.save()
    run = ln.Run(transform=transform)
    assert _add_emoji(run, label="run") == "🖥️ run"
    transform.delete(permanent=True)


def test_view_lineage_circular():
    import pandas as pd

    transform = ln.Transform(key="test").save()
    run = ln.Run(transform=transform).save()
    artifact = ln.Artifact.from_dataframe(
        pd.DataFrame({"a": [1, 2, 3]}), description="test artifact", run=run
    ).save()
    run.input_artifacts.add(artifact)
    artifact.view_lineage()
    artifact.delete(permanent=True)
    transform.delete(permanent=True)



import shutil
from pathlib import Path
from typing import Generator

import anndata as ad
import lamindb as ln
import mudata as md
import numpy as np
import pandas as pd
import pytest
import spatialdata as sd
import tiledbsoma
import tiledbsoma.io
from scipy.sparse import csr_matrix


@pytest.fixture(scope="session")
def get_small_adata():
    return ad.AnnData(
        X=np.array([[1, 2, 3], [4, 5, 6]]),
        obs={"feat1": ["A", "B"]},
        var=pd.DataFrame(index=["MYC", "TCF7", "GATA1"]),
        obsm={"X_pca": np.array([[1, 2], [3, 4]])},
    )


@pytest.fixture(scope="session")
def get_small_mdata():
    adata1 = ad.AnnData(
        X=np.array([[1, 2, 3], [4, 5, 6]]),
        obs={"feat1": ["A", "B"]},
        var=pd.DataFrame(index=["MYC", "TCF7", "GATA1"]),
        obsm={"X_pca": np.array([[1, 2], [3, 4]])},
    )

    adata2 = ad.AnnData(
        X=np.array([[7, 8], [9, 10]]),
        obs={"feat2": ["C", "D"]},
        var=pd.DataFrame(index=["FOXP3", "CD8A"]),
        obsm={"X_umap": np.array([[5, 6], [7, 8]])},
    )

    return md.MuData({"rna": adata1, "protein": adata2})


@pytest.fixture(scope="session")
def get_small_sdata():
    adata = ad.AnnData(
        X=csr_matrix(np.array([[0.1, 0.2], [0.3, 0.4]])),
        obs=pd.DataFrame(index=["cell1", "cell2"]),
        var=pd.DataFrame(index=["gene1", "gene2"]),
    )

    {
        "region1": np.array([[[0, 0], [0, 1], [1, 1], [1, 0]]]),
        "region2": np.array([[[2, 2], [2, 3], [3, 3], [3, 2]]]),
    }

    sdata_obj = sd.SpatialData(
        tables={"gene_expression": adata},
    )

    return sdata_obj


@pytest.fixture(scope="session")
def get_small_soma_experiment():
    adata = ln.core.datasets.mini_immuno.get_dataset1(otype="AnnData")
    tiledbsoma.io.from_anndata("test.tiledbsoma", adata, measurement_name="RNA")

    exp = tiledbsoma.Experiment.open("test.tiledbsoma")
    yield exp

    shutil.rmtree("test.tiledbsoma")


@pytest.fixture(scope="session")
def get_mini_csv() -> Generator[Path, None, None]:
    csv_path = ln.core.datasets.file_mini_csv()
    yield csv_path

    Path("mini.csv").unlink(missing_ok=True)



import lamindb as ln
import bionty as bt
import tiledbsoma as soma
import tiledbsoma.io

adata = ln.core.datasets.mini_immuno.get_dataset1(otype="AnnData")
tiledbsoma.io.from_anndata("small_dataset.tiledbsoma", adata, measurement_name="RNA")

obs_schema = ln.Schema(
    name="soma_obs_schema",
    features=[
        ln.Feature(name="cell_type_by_expert", dtype=bt.CellType).save(),
        ln.Feature(name="cell_type_by_model", dtype=bt.CellType).save(),
    ],
).save()

var_schema = ln.Schema(
    name="soma_var_schema",
    features=[
        ln.Feature(name="var_id", dtype=bt.Gene.ensembl_gene_id).save(),
    ],
    coerce_dtype=True,
).save()

soma_schema = ln.Schema(
    name="soma_experiment_schema",
    otype="tiledbsoma",
    slots={
        "obs": obs_schema,
        "ms:RNA.T": var_schema,
    },
).save()

with soma.Experiment.open("small_dataset.tiledbsoma") as experiment:
    curator = ln.curators.TiledbsomaExperimentCurator(experiment, soma_schema)
    curator.validate()
    artifact = curator.save_artifact(
        key="examples/soma_experiment.tiledbsoma",
        description="SOMA experiment with schema validation",
    )
assert artifact.schema == soma_schema
artifact.describe()



# Generated by Django 5.2 on 2025-01-27 13:48

from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0084_alter_schemafeature_feature_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="feature",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="flextable",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="param",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
    ]



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from ._anndata_accessor import AnnDataAccessor

if TYPE_CHECKING:
    from zarr import Group

    from lamindb import Artifact


class _TablesAccessor:
    def __init__(self, tables: Group, artifact: Artifact | None = None):
        self._tables = tables

        self._artifact = artifact

    def __getitem__(self, key: str) -> AnnDataAccessor:
        return AnnDataAccessor(
            connection=None,
            storage=self._tables[key],
            filename=key,
            artifact=self._artifact,
        )

    def keys(self) -> list[str]:
        return list(self._tables.keys())

    def __repr__(self) -> str:
        """Description of the _TablesAccessor object."""
        descr = (
            f"Accessor for the SpatialData attribute tables\n  with keys: {self.keys()}"
        )
        return descr


class SpatialDataAccessor:
    """Cloud-backed SpatialData.

    For now only allows to access `tables`.
    """

    def __init__(self, storage: Group, name: str, artifact: Artifact | None = None):
        self.storage = storage
        self._name = name

        self._artifact = artifact

    @cached_property
    def tables(self) -> _TablesAccessor:
        """tables of the underlying SpatialData object."""
        return _TablesAccessor(self.storage["tables"], self._artifact)

    def __repr__(self):
        """Description of the SpatialDataAccessor object."""
        descr = (
            "SpatialDataAccessor object"
            f"\n  constructed for the SpatialData object {self._name}"
            f"\n    with tables: {self.tables.keys()}"
        )
        return descr



import lamindb as ln
import bionty as bt

obs_schema = ln.examples.schemas.valid_features()
varT_schema = ln.Schema(
    name="valid_ensembl_gene_ids", itype=bt.Gene.ensembl_gene_id
).save()
schema = ln.Schema(
    name="anndata_ensembl_gene_ids_and_valid_features_in_obs",
    otype="AnnData",
    slots={"obs": obs_schema, "var.T": varT_schema},
).save()



# Generated by Django 5.2 on 2025-05-07 12:16

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0092_alter_artifactfeaturevalue_artifact_and_more"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="schemacomponent",
            unique_together={("composite", "slot"), ("composite", "slot", "component")},
        ),
    ]



import bionty as bt
import lamindb as ln
import pandas as pd
import pytest


@pytest.fixture(scope="module")
def df():
    return pd.DataFrame(
        (
            ["T cell", "CL:0000084"],
            ["hepatocyte", "CL:0000182"],
            ["my new cell type", ""],
        ),
        columns=["cell_type", "cell_type_id"],
    )


def test_from_values_name(df):
    bt.CellType.filter().delete()
    assert df["cell_type"].tolist() == ["T cell", "hepatocyte", "my new cell type"]
    # create records from bionty
    result = bt.CellType.from_values(df.cell_type, "name")
    ids = [i.ontology_id for i in result]
    assert len(result) == 2
    assert set(ids) == {"CL:0000084", "CL:0000182"}
    assert result[0].source.entity == "bionty.CellType"

    # wrong field type
    with pytest.raises(TypeError):
        result = bt.CellType.from_values(df.cell_type, field=bt.CellType)


def test_from_values_ontology_id(df):
    assert df["cell_type_id"].tolist() == ["CL:0000084", "CL:0000182", ""]
    result = bt.CellType.from_values(df.cell_type_id, "ontology_id")
    names = {i.name for i in result}
    assert len(result) == 2
    assert names == {"T cell", "hepatocyte"}
    assert result[0].source.entity == "bionty.CellType"


def test_from_values_multiple_match():
    records = bt.Gene.from_values(["ABC1", "PDCD1"], bt.Gene.symbol, organism="human")
    assert len(records) == 3


def test_get_or_create_records():
    names = ["ulabel" + str(i) for i in range(25)]
    labels = [ln.ULabel(name=name) for name in names]
    ln.save(labels)
    # more than 20 existing values
    labels = ln.ULabel.from_values(names, field="name")
    assert len(labels) == 25


def test_from_values_synonyms_aware():
    bt.CellType.from_source(name="T cell").save()
    # existing validated values
    records = bt.CellType.from_values(["T cell"], "name")
    assert len(records) == 1
    assert records[0].name == "T cell"
    assert isinstance(records[0].source, bt.Source)
    # existing validated values and synonyms
    records = bt.CellType.from_values(["T cell", "T-cell"], "name")
    assert len(records) == 1
    assert records[0].name == "T cell"
    assert isinstance(records[0].source, bt.Source)
    # bionty values and synonyms
    records = bt.CellType.from_values(["B-cell", "B cell"], "name")
    assert len(records) == 1
    assert records[0].name == "B cell"
    assert isinstance(records[0].source, bt.Source)
    # all possibilities of validated values
    records = bt.CellType.from_values(
        ["T cell", "T-cell", "t cell", "B cell", "B-cell"], "name"
    )
    assert len(records) == 2
    names = [r.name for r in records]
    assert set(names) == {"T cell", "B cell"}
    assert isinstance(records[0].source, bt.Source)
    assert isinstance(records[1].source, bt.Source)
    # non-validated values
    records = bt.CellType.from_values(["T cell", "mycell"], "name")
    assert len(records) == 1
    assert records[0].name == "T cell"
    assert isinstance(records[0].source, bt.Source)
    assert records[0].ontology_id == "CL:0000084"
    bt.CellType.filter().delete()


def test_standardize():
    # only name field can be standardized
    results = bt.Gene.from_values(
        ["HES4", "TNFRSF4"], field=bt.Gene.ensembl_gene_id, organism="human"
    )
    assert len(results) == 0

    results = bt.Gene.from_values(
        ["HES4", "TNFRSF4"], field=bt.Gene.symbol, organism="human"
    )
    assert len(results) == 2
